"""network_service.py

This module contains functions and classes for the creation of the
VegaServiceNetwork class. A VegaServiceNetwork object can be used to
communicate with either a public or internal Vega network and run a
Vega Wallet service for communication with the chosen Vega network.
The VegaServiceNetwork inherits properties and methods from the VegaService
class. Inherited methods can be used to communicate with the Vega datanode
and Vega wallet services. Redundant properties and methods are overwritten.

A vegawallet executable should either be in PATH whilst executing, or 
the VEGA_WALLET_PATH environment variable set to a location of a wallet.

Similarly, specify VEGA_NETWORK_CONFIG for the path to the network config
of your chosen network

Example:

    For an example, try running the below command. It will create a
    VegaServiceNetwork object, the relevant services and try sending some
    commands.

        $ python -m vega_sim.network_service

    If the above example is throwing errors, check you have cloned the
    required bin folders with the following command.

        $ make networks

Todo:
    • Test all public and internal Vega networks are compatible.
    • Test all market sim agents are compatible.
"""

import grpc
import toml
import time
import signal
import logging
import tempfile
import itertools
import subprocess
import webbrowser
import multiprocessing
import requests

from typing import Optional

from os import getcwd, path, environ

import vega_sim.grpc.client as vac

from vega_sim import vega_bin_path
from vega_sim.service import VegaService, DatanodeBehindError, DatanodeSlowResponseError
from vega_sim.wallet.base import Wallet
from vega_sim.wallet.vega_wallet import VegaWallet
from vega_sim.null_service import find_free_port, _popen_process

from vega_sim.constants import DATA_NODE_GRPC_PORT
from vega_sim.scenario.constants import Network

from vega_sim.api.helpers import num_to_padded_int, num_from_padded_int

logger = logging.getLogger(__name__)


class VegaWalletStartupTimeoutError(Exception):
    pass


def add_network_config(
    network_config_path: str,
):
    """Creates a network config file for the specified network.
    Method creates a copy of the config file specified in the local vega path
    networks folder. This config file is used to run a wallet service.

    Args:
        network_config_path (str):
            Path to the network config file to create a local network from.

    Raises:
        ValueError:
            If no file exists at the specified path.
    """

    if not path.exists(network_config_path):
        raise ValueError("No network config file at the specified path.")

    vega_wallet_path = environ.get("VEGA_WALLET_PATH", "vegawallet")

    args = [
        vega_wallet_path,
        "network",
        "import",
        "--from-file",
        network_config_path,
        "--force",
    ]

    subprocess.call(args=args)


def manage_vega_processes(
    network: str,
    vega_console_path: str,
    vega_console_port: int,
    data_node_query_url: str,
    wallet_url: str,
    run_with_wallet: Optional[bool] = True,
    run_with_console: Optional[bool] = True,
    log_dir: Optional[str] = None,
):
    processes = []

    tmp_vega_dir = tempfile.mkdtemp(prefix="vega-sim-") if log_dir is None else log_dir

    if run_with_wallet:
        vega_wallet_path = environ.get("VEGA_WALLET_PATH", "vegawallet")
        vegaWalletProcess = _popen_process(
            popen_args=[
                vega_wallet_path,
                "service",
                "run",
                "--network",
                network,
                "--automatic-consent",
                "--load-tokens",
                "--tokens-passphrase-file",
                environ.get("VEGA_WALLET_TOKENS_PASSPHRASE_FILE"),
                "--no-version-check",
            ],
            dir_root=tmp_vega_dir,
            log_name="vegawallet",
        )
        processes.append(vegaWalletProcess)

    if run_with_console:
        env_copy = environ.copy()
        env_copy.update(
            {
                "NX_VEGA_URL": data_node_query_url,
                "NX_VEGA_WALLET_URL": wallet_url,
                "NX_VEGA_ENV": "CUSTOM",
                "NX_PORT": f"{vega_console_port}",
            }
        )

        vegaConsoleProcess = _popen_process(
            popen_args=[
                "yarn",
                "--cwd",
                vega_console_path,
                "nx",
                "serve",
                "-o",
                "trading",
                "--port",
                f"{vega_console_port}",
            ],
            dir_root=tmp_vega_dir,
            log_name="console",
            env=env_copy,
        )
        processes.append(vegaConsoleProcess)

    signal.sigwait([signal.SIGKILL, signal.SIGTERM])
    for process in processes:
        process.terminate()
    for process in processes:
        return_code = process.poll()
        if return_code is not None:
            continue
        time.sleep(5)
        process.kill()


def _find_network_config_toml(
    network: Network, config_path: Optional[str] = None
) -> Optional[str]:
    search_paths = (
        [config_path]
        if config_path is not None
        else [
            path.join(
                getcwd(),
                "vega_sim",
                "bin",
                "networks-internal",
                network.name.lower(),
            ),
            path.join(
                getcwd(),
                "vega_sim",
                "bin",
                "networks",
                network.name.lower(),
            ),
        ]
    )
    for search_path in search_paths:
        full_path = path.join(
            search_path,
            f"{network.value}.toml",
        )

        if path.exists(full_path):
            return full_path


class VegaServiceNetwork(VegaService):
    """Class for handling services for communicating with a Vega network."""

    def __init__(
        self,
        network: Network,
        run_with_wallet: bool = True,
        run_with_console: bool = True,
        vega_console_path: Optional[str] = None,
        network_config_path: Optional[str] = None,
        wallet_path: Optional[str] = None,
        vega_home_path: Optional[str] = None,
        wallet_home_path: Optional[str] = None,
        wallet_token_path: Optional[str] = None,
        wallet_passphrase_path: Optional[str] = None,
        wallet_url: Optional[bool] = None,
        faucet_url: Optional[bool] = None,
        vega_node_grpc_url: Optional[str] = None,
        load_existing_keys: bool = True,
        governance_symbol: Optional[str] = "VEGA",
        vegacapsule_bin_path: Optional[str] = "./vega_sim/bin/vegacapsule",
        network_on_host: Optional[bool] = False,
    ):
        """Method initialises the class.

        Args:
            network (Network):
                Defines which network to connect service to.
            run_with_wallet (bool, optional):
                Defines whether to start a wallet process.
            run_with_console (bool, optional):
                Defines whether to start a console process.
            vega_console_path (str, optional):
                Path to the directory containing console files if
                wishing to run a local console
            network_config_path (str, optional):
                Path to the directory containing network config files.
                If not passed will search first the environment variable
                VEGA_NETWORK_CONFIG then two default paths.
                Note: Only needed if creating keys
            vega_home_path (str, optional):
                Path to the directory containing wallet binary. Otherwise
                uses VEGA_HOME environment variable.
                Note: Only needed if creating keys
            wallet_token_path (str, optional):
                Path to the json file containing wallet tokens. Otherwise
                uses VEGA_WALLET_TOKENS_FILE environment variable.
        """

        # Run init method inherited from VegaService with network arguments.
        super().__init__(
            can_control_time=False,
            warn_on_raw_data_access=False,
            governance_symbol=governance_symbol,
        )

        self.network = network
        self.run_with_wallet = run_with_wallet
        self.run_with_console = run_with_console

        self._wallet = None
        self._data_node_grpc_url = None
        self._data_node_query_url = None
        self._data_node_rest_url = None

        self._wallet_url = wallet_url
        self._faucet_url = faucet_url

        self._network_config = None

        self.vega_console_path = (
            vega_console_path
            if vega_console_path is not None
            else path.join(vega_bin_path, "console")
        )
        self._base_network_config_path = (
            network_config_path
            if network_config_path is not None
            else environ.get("VEGA_NETWORK_CONFIG")
        )

        self._wallet_path = (
            wallet_path
            if wallet_path is not None
            else environ.get("VEGA_WALLET_PATH", "vegawallet")
        )

        self._vega_home = (
            vega_home_path if vega_home_path is not None else environ.get("VEGA_HOME")
        )
        self._wallet_home = (
            wallet_home_path
            if wallet_home_path is not None
            else environ.get("WALLET_HOME")
        )

        self._passphrase_file_path = (
            wallet_passphrase_path
            if wallet_passphrase_path is not None
            else environ.get("VEGA_WALLET_TOKENS_PASSPHRASE_FILE")
        )

        self._token_path = (
            wallet_token_path
            if wallet_token_path is not None
            else environ.get("VEGA_WALLET_TOKENS_FILE")
        )

        self._network_config_path = _find_network_config_toml(
            network=self.network, config_path=self._base_network_config_path
        )
        if self._network_config_path is None:
            raise ValueError(
                f"ERROR! {self.network.name.lower()} network config could not be found"
            )
        self._grpc_endpoints = itertools.cycle(
            self.network_config["API"]["GRPC"]["Hosts"]
        )
        self._rest_endpoints = itertools.cycle(
            self.network_config["API"]["REST"]["Hosts"]
        )
        self._graphql_endpoints = itertools.cycle(
            self.network_config["API"]["GraphQL"]["Hosts"]
        )

        self.load_existing_keys = load_existing_keys
        self.vegacapsule_bin_path = vegacapsule_bin_path

        self.log_dir = tempfile.mkdtemp(prefix="vega-sim-")

        self._vega_node_grpc_url = vega_node_grpc_url

        if network_on_host:
            logging.info(
                "Network running on host machine. Updating url to use host.docker.internal"
            )

            def update_endpoint(endpoint, http: bool = True):
                return (
                    ("http://" if http else "")
                    + "host.docker.internal:"
                    + endpoint.split(":")[-1]
                )

            def update_iterator(iterator, http: bool = True):
                for endpoint in iterator:
                    yield update_endpoint(endpoint, http)

            self._grpc_endpoints = update_iterator(self._grpc_endpoints, http=False)
            self._rest_endpoints = update_iterator(self._rest_endpoints)
            self._graphql_endpoints = update_iterator(self._graphql_endpoints)
            self._wallet_url = update_endpoint(self._wallet_url)
            self._faucet_url = update_endpoint(self._faucet_url)

            if self._vega_node_grpc_url is not None:
                self._vega_node_grpc_url = update_endpoint(
                    self._vega_node_grpc_url, http=False
                )

    def __enter__(self):
        """Defines behaviour when class entered by a with statement."""
        self.start()

        return self

    def __exit__(self, type, value, traceback):
        """Defines behaviour when class exited by a with statement."""
        self.stop()

    def start(self):
        ctx = multiprocessing.get_context()
        vega_console_port = find_free_port()
        self.proc = ctx.Process(
            target=manage_vega_processes,
            kwargs={
                "network": self.network.name.lower(),
                "log_dir": self.log_dir,
                "run_with_wallet": self.run_with_wallet,
                "run_with_console": self.run_with_console,
                "vega_console_path": self.vega_console_path,
                "vega_console_port": vega_console_port,
                "data_node_query_url": self.data_node_query_url,
                "wallet_url": self.wallet_url,
            },
        )
        self.proc.start()

        if self.run_with_console:
            logger.info(
                "Vega Running. Console launched at"
                f" http://localhost:{vega_console_port}"
            )
            webbrowser.open(f"http://localhost:{vega_console_port}/", new=2)

        started = False
        for _ in range(3600):
            try:
                response = requests.get(f"{self.wallet_url}/api/v2/health")
                response.raise_for_status()
                started = True
                break
            except (
                requests.exceptions.ConnectionError,
                requests.exceptions.HTTPError,
            ):
                time.sleep(0.1)
        if not started:
            raise VegaWalletStartupTimeoutError(
                "Timed out waiting for Vega wallet service"
            )

        self.check_datanode(raise_on_error=False)

    def stop(self) -> None:
        if self.proc is None:
            logger.info("Stop called but nothing to stop")
        else:
            self.proc.terminate()
        super().stop()

    def wait_fn(self, wait_multiple: float = 1) -> None:
        """Overrides redundant parent method."""
        logging.debug(
            "Parent method overridden as VegaNetworkService incapable of controlling"
            " time."
        )

    def wait_for_datanode_sync(self) -> None:
        """Overrides redundant parent method."""
        logging.debug(
            "Parent method overridden as VegaNetworkService incapable of controlling"
            " time."
        )

    def wait_for_core_catchup(self) -> None:
        """Overrides redundant parent method."""
        logging.debug(
            "Parent method overridden as VegaNetworkService incapable of controlling"
            " time."
        )

    def wait_for_total_catchup(self) -> None:
        """Overrides redundant parent method."""
        logging.debug(
            "Parent method overridden as VegaNetworkService incapable of controlling"
            " time."
        )

    @property
    def network_config(self) -> dict:
        if self._network_config is None:
            self._network_config = toml.load(self._network_config_path)
            if self.run_with_wallet:
                add_network_config(self._network_config_path)
        return self._network_config

    @property
    def data_node_grpc_url(self) -> str:
        if self._data_node_grpc_url is None:
            self._data_node_grpc_url = next(self._grpc_endpoints)
        return self._data_node_grpc_url

    @property
    def data_node_rest_url(self) -> str:
        if self._data_node_rest_url is None:
            self._data_node_rest_url = next(self._rest_endpoints)
        return self._data_node_rest_url

    @property
    def data_node_query_url(self) -> str:
        if self._data_node_query_url is None:
            self._data_node_query_url = next(self._graphql_endpoints)
        return self._data_node_query_url

    @property
    def vega_node_grpc_url(self) -> str:
        return self._vega_node_grpc_url

    @property
    def wallet_url(self) -> str:
        if self._wallet_url is None:
            self._wallet_url = f"http://127.0.0.1:1789"
        return self._wallet_url

    @property
    def faucet_url(self) -> str:
        return self._faucet_url

    @property
    def wallet(self) -> Wallet:
        if self._wallet is None:
            if self.load_existing_keys:
                if self._token_path is None:
                    raise Exception(
                        "Either path to tokens JSON must be passed to wallet class or"
                        " VEGA_WALLET_TOKENS_FILE environment variable set"
                    )
                self._wallet = VegaWallet.from_json(
                    self._token_path,
                    self.wallet_url,
                    wallet_path=self._wallet_path,
                    vega_home_dir=self._vega_home,
                )
            else:
                self._wallet = VegaWallet(
                    wallet_url=self.wallet_url,
                    wallet_path=self._wallet_path,
                    vega_home_dir=self._wallet_home,
                    passphrase_file_path=self._passphrase_file_path,
                )
        return self._wallet

    @property
    def core_state_client(self) -> None:
        logging.debug(
            (
                "Parent property overridden as VegaNetworkService does not need a core"
                " client."
            ),
        )
        pass

    def check_datanode(
        self, max_time_diff: int = 30, raise_on_error: Optional[bool] = True
    ):
        """Checks if the current data-node connection is healthy.

        If the current data-node connection has timed out or the end-point is
        unresponsive then the service will attempt to establish a connection
        with a new healthy data-node.

        Args:
            max_time_diff (int, optional):
                Maximum allowable difference between system time and datanode time in
                seconds. Defaults to 30.
            raise_on_error (bool, optional):
                Whether to raise an error if data-node connection unhealthy.
        """

        try:
            self.ping_datanode(max_time_diff=max_time_diff)
            logging.debug(
                f"Connection to endpoint {self._data_node_grpc_url} successful."
            )
            return

        except grpc.FutureTimeoutError as e:
            if raise_on_error:
                raise e
            else:
                logging.warning(
                    f"Connection to endpoint {self._data_node_grpc_url} timed out."
                )
                self.switch_datanode()

        except grpc._channel._InactiveRpcError as e:
            if raise_on_error:
                raise e
            else:
                logging.warning(
                    f"Connection to endpoint {self._data_node_grpc_url} inactive."
                )
                self.switch_datanode()

        except DatanodeBehindError as e:
            if raise_on_error:
                raise e
            else:
                logging.warning(
                    f"Connection to endpoint {self._data_node_grpc_url} is behind."
                )
                self.switch_datanode()

        except DatanodeSlowResponseError as e:
            if raise_on_error:
                raise e
            else:
                logging.warning(
                    f"Connection to endpoint {self._data_node_grpc_url} is slow."
                )
                self.switch_datanode()

    def switch_datanode(self, max_attempts: Optional[int] = -1):
        """Attempts to establish a new data-node connection.

        Args:
            max_attempts (int, optional):
                Maximum number of connection attempts to attempt before raising an
                error. Defaults to -1 (infinite attempts).
        """

        attempts = 0
        while True:
            attempts += 1

            try:
                self._data_node_grpc_url = next(self._grpc_endpoints)

                logging.debug(f"Switched to endpoint {self._data_node_grpc_url}")

                channel = grpc.insecure_channel(
                    self.data_node_grpc_url,
                    options=(
                        ("grpc.enable_http_proxy", 0),
                        ("grpc.max_send_message_length", 1024 * 1024 * 20),
                        ("grpc.max_receive_message_length", 1024 * 1024 * 20),
                    ),
                )
                grpc.channel_ready_future(channel).result(timeout=30)
                self._trading_data_client_v2 = vac.VegaTradingDataClientV2(
                    self.data_node_grpc_url,
                    channel=channel,
                )

                # Ping the datanode to check it is not behind
                self.ping_datanode()
                logging.debug(
                    f"Connection to endpoint {self.data_node_grpc_url} successful."
                )

                return

            except grpc._channel._InactiveRpcError:
                logging.warning(
                    f"Connection to endpoint {self.data_node_grpc_url} inactive."
                )

            except grpc.FutureTimeoutError:
                logging.warning(
                    f"Connection to endpoint {self.data_node_grpc_url} timed out."
                )

            except DatanodeBehindError:
                logging.warning(
                    f"Connection to endpoint {self.data_node_grpc_url} is behind."
                )

            except DatanodeSlowResponseError:
                logging.warning(
                    f"Connection to endpoint {self.data_node_grpc_url} is slow."
                )

            if attempts == max_attempts:
                break

        raise Exception("Unable to establish connection to a data-node.")

    def mint(
        self,
        key_name: Optional[str],
        asset: str,
        amount: float,
        wallet_name: Optional[str] = None,
    ) -> None:
        """Mints a given amount of requested asset into the associated wallet

        Args:
            wallet_name:
                str, The name of the wallet
            asset:
                str, The ID of the asset to mint
            amount:
                float, the amount of asset to mint
            key_name:
                Optional[str], key name stored in metadata. Defaults to None.
        """
        details = self.get_asset(asset).details
        is_erc20 = True if details.erc20.contract_address != "" else False

        if is_erc20:
            self.deposit(
                symbol=details.symbol,
                amount=amount,
                key_name=key_name,
                wallet_name=wallet_name,
            )
        else:
            max_faucet_amount = num_from_padded_int(
                details.builtin_asset.max_faucet_amount_mint, details.decimals
            )
            super().mint(
                key_name=key_name,
                asset=asset,
                amount=amount if amount < max_faucet_amount else max_faucet_amount,
                wallet_name=wallet_name,
            )

    def deposit(
        self,
        symbol: str,
        amount: float,
        key_name: str,
        wallet_name: Optional[str] = None,
    ):
        """Deposit an amount of a specified ERC20 asset to a a vega key.

        Args:
            symbol (str): Symbol of the asset to be deposited.
            amount (float): Amount to be deposited.
            key_name (str): Name of the key used for the deposit.
            wallet_name (Optional[str], optional): Name of the wallet. Defaults to the
                default wallet_name set in the wallet.

        Raises:
            Exception: Raised if the deposit does not arrive within a fixed time limit.
        """

        asset_id = self.find_asset_id(symbol=symbol)
        amount = str(
            num_to_padded_int(
                amount,
                self.asset_decimals[asset_id],
            )
        )

        pub_key = self.wallet.public_key(name=key_name, wallet_name=wallet_name)
        account_before = self.party_account(party_id=pub_key, asset_id=asset_id).general

        args = [
            self.vegacapsule_bin,
            "ethereum",
            "asset",
            "deposit",
            "--asset-symbol",
            symbol,
            "--amount",
            amount,
            "--pubkey",
            pub_key,
        ]
        subprocess.run(args)

        for _ in range(500):
            if self.party_account(party_id=pub_key, asset_id=asset_id) > account_before:
                return
            self.wait_fn(1)
        raise Exception("Deposit never arrived.")

    def stake(
        self,
        amount: float,
        key_name: str,
        wallet_name: Optional[str] = None,
    ):
        """Stake a specified amount of a governance asset.

        Args:
            amount (float): The amount to stake.
            key_name (str): The name of the key used for staking.
            wallet_name (Optional[str], optional): The name of the wallet. Defaults to
                the default value set in the wallet.

        Raises:
            Exception: Raised if the stake does not arrive within a given time limit.
        """
        amount = str(
            num_to_padded_int(
                amount,
                self.asset_decimals[self.find_asset_id(symbol=self.governance_symbol)],
            )
        )

        pub_key = self.wallet.public_key(name=key_name, wallet_name=wallet_name)
        stake_before = self.get_stake(party_id=pub_key)

        args = [
            self.vegacapsule_bin_path,
            "ethereum",
            "asset",
            "stake",
            "--amount",
            amount,
            "--asset-symbol",
            self.governance_symbol,
            "--pub-key",
            pub_key,
            "--home-path",
            self._vega_home,
        ]
        subprocess.run(args)

        for _ in range(500):
            if self.get_stake(party_id=pub_key) > stake_before:
                return
            self.wait_fn(1)
        raise Exception("Stake never arrived.")

    def wait_fn(self, wait_multiple: float = 1) -> None:
        time.sleep(wait_multiple)


if __name__ == "__main__":
    """Module Example"""

    log = logging.basicConfig(level=logging.INFO)

    # Create a service connected to the fairground network.
    with VegaServiceNetwork(
        network=Network.FAIRGROUND,
        run_with_wallet=True,
        run_with_console=True,
    ) as vega:
        # Show all the markets on the network
        markets = vega.all_markets()
        logging.info(markets)

        # Show data for a specific market
        market = vega.get_latest_market_data(market_id=markets[0].id)

    # Create a service connected to the stagnet3 network.
    with VegaServiceNetwork(
        network=Network.FAIRGROUND,
        run_with_wallet=True,
        run_with_console=True,
    ) as vega:
        # Show all the markets on the network
        markets = vega.all_markets()
        logging.info(markets)

        # Show data for a specific market
        market = vega.get_latest_market_data(market_id=markets[0].id)



from abc import abstractmethod
from collections import defaultdict, namedtuple
from typing import Dict, List, Optional, Tuple

import numpy as np

from vega_sim.environment import VegaState
from vega_sim.environment.agent import Agent, StateAgentWithWallet
from vega_sim.null_service import VegaServiceNull
from vega_sim.proto.vega import markets as markets_protos

from vega_sim.reinforcement.la_market_state import (
    AbstractAction,
    LAMarketState,
    states_to_sarsa,
)

WalletConfig = namedtuple("WalletConfig", ["name", "passphrase"])

# Send selling/buying MOs to hit LP orders
WALLET = WalletConfig("learner", "learner")


def state_fn(
    service: VegaServiceNull,
    agents: Dict[str, Agent],
    state_values=None,
) -> Tuple[LAMarketState, AbstractAction]:
    learner = agents["learner"]
    return (learner.latest_state, learner.latest_action)


class LearningAgent(StateAgentWithWallet):
    @abstractmethod
    def __init__(
        self,
        device: str,
        logfile_pol_imp: str,
        logfile_pol_eval: str,
        logfile_pnl: str,
        discount_factor: float,
        num_levels: int,
        key_name: str,
        market_name: str,
        initial_balance: int,
        position_decimals: int,
        asset_name: str,
        inventory_penalty: float = 0.0,
    ):
        super().__init__(key_name=key_name)

        self.step_num = 0
        self.latest_state = None
        self.latest_action = None
        self.device = device
        self.discount_factor = discount_factor
        self.initial_balance = initial_balance

        self.memory = defaultdict(list)
        self.memory_capacity = 100_000

        # Coefficients for regularisation
        self.coefH_discr = 1.0
        self.coefH_cont = 0.01
        # losses logger
        self.losses = defaultdict(list)
        # logfile
        self.logfile_pol_imp = logfile_pol_imp
        self.logfile_pol_eval = logfile_pol_eval
        self.logfile_pnl = logfile_pnl

        self.lerningIteration = 0
        self.market_name = market_name
        self.position_decimals = position_decimals
        self.inventory_penalty = inventory_penalty
        self.asset_name = asset_name

    def set_market_tag(self, tag: str):
        self.tag = tag

    @abstractmethod
    def move_to_device(self):
        pass

    @abstractmethod
    def move_to_cpu(self):
        pass

    def learning_step(self, results_dir: Optional[str] = None):
        # Policy evaluation + Policy improvement
        self.move_to_device()
        self.policy_eval(batch_size=20000, n_epochs=10)
        self.policy_improvement(batch_size=100_000, n_epochs=10)

        if results_dir is not None:
            # save in case environment chooses to crash
            self.save(results_dir)
        self.move_to_cpu()

    def initialise(self, vega: VegaServiceNull):
        # Initialise wallet
        super().initialise(vega=vega)
        self.step_num = 0
        market_name = self.market_name + f"_{self.tag}"

        # Get market id
        self.market_id = [
            m.id
            for m in self.vega.all_markets()
            if m.tradable_instrument.instrument.name == market_name
        ][0]
        # Get asset id
        self.tdai_id = self.vega.find_asset_id(symbol=self.asset_name)
        # Top up asset
        self.vega.mint(
            self.key_name,
            asset=self.tdai_id,
            amount=self.initial_balance,
        )
        self.vega.wait_fn(2)

    def _update_memory(
        self,
        state: LAMarketState,
        action: AbstractAction,
        reward: float,
        next_state: LAMarketState,
    ):
        pass

    def update_memory(self, states: List[Tuple[LAMarketState, AbstractAction]]):
        """
        Updates memory of the agent, and removes old tuples (s,a,r,s) if memory exceeds its capacity
        """
        for res in states_to_sarsa(states, inventory_penalty=self.inventory_penalty):
            self._update_memory(res[0], res[1], res[2], res[3])
        # remove old tuples if memory exceeds its capaciy
        for key, value in self.memory.items():
            if len(self.memory[key]) > self.memory_capacity:
                first_n = len(self.memory[key]) - self.memory_capacity
                del self.memory[key][:first_n]
        return 0

    def clear_memory(self):
        for key in self.memory.keys():
            self.memory[key].clear()

    @abstractmethod
    def create_dataloader(self, batch_size):
        """
        creates dataset and dataloader for training.
        """
        pass

    def state(self, vega: VegaServiceNull) -> LAMarketState:
        position = self.vega.positions_by_market(self.key_name, self.market_id)

        position = position.open_volume if position else 0
        account = self.vega.party_account(
            key_name=self.key_name,
            asset_id=self.tdai_id,
            market_id=self.market_id,
        )
        book_state = self.vega.market_depth(
            self.market_id, num_levels=self.num_levels
        )  # make num_levels as a parameter?

        market_info = vega.market_info(market_id=self.market_id)
        fee = (
            float(market_info.fees.factors.liquidity_fee)
            + float(market_info.fees.factors.maker_fee)
            + float(market_info.fees.factors.infrastructure_fee)
        )
        init_price = self.price_process[0]
        next_price = (
            self.price_process[self.step_num + 1]
            if self.price_process is not None
            and len(self.price_process) > self.step_num + 1
            else np.nan
        )
        next_price /= init_price
        bid_prices = [level.price / init_price for level in book_state.buys] + [
            0
        ] * max(0, self.num_levels - len(book_state.buys))
        ask_prices = [level.price / init_price for level in book_state.sells] + [
            0
        ] * max(0, self.num_levels - len(book_state.sells))

        return LAMarketState(
            step=self.step_num,
            position=position,
            full_balance=(account.margin + account.general) / self.initial_balance,
            market_in_auction=(
                not market_info.trading_mode
                == markets_protos.Market.TradingMode.TRADING_MODE_CONTINUOUS
            ),
            bid_prices=bid_prices,
            ask_prices=ask_prices,
            bid_volumes=[level.volume for level in book_state.buys]
            + [0] * max(0, self.num_levels - len(book_state.buys)),
            ask_volumes=[level.volume for level in book_state.sells]
            + [0] * max(0, self.num_levels - len(book_state.sells)),
            trading_fee=fee,
            next_price=next_price,
        )

    @abstractmethod
    def empty_action(self) -> AbstractAction:
        pass

    def finalise(self):
        numTries = 3
        account = None
        for i in range(0, numTries):
            account = self.vega.party_account(
                key_name=self.key_name,
                asset_id=self.tdai_id,
                market_id=self.market_id,
            )
            self.latest_state = self.state(self.vega)
            if account.margin == 0:
                break
            self.vega.forward("1s")
            self.vega.wait_for_total_catchup()

        if account.margin > 0:
            print(
                "Market should be settled but there is still balance in margin account."
                " What's up?"
            )

        self.latest_action = self.empty_action()
        self.step_num += 1
        # final_pnl = self.latest_state.full_balance - self.initial_balance
        final_pnl = self.latest_state.full_balance - 1.0
        with open(self.logfile_pnl, "a") as f:
            f.write("{},{:.8f}\n".format(self.lerningIteration, final_pnl))

        return super().finalise()

    @abstractmethod
    def step(self, vega_state: VegaState):
        pass

    @abstractmethod
    def save(self, results_dir: str):
        pass

    @abstractmethod
    def load(self, results_dir: str):
        pass



from dataclasses import dataclass
from typing import Optional, List

from vega_sim.scenario.common.utils.price_process import Granularity


MAX_FAUCET = 1e10


@dataclass
class MarketManagerArgs:
    market_name: str
    market_code: str
    asset_name: str
    adp: int
    mdp: int
    pdp: int
    max_faucet: Optional[int] = MAX_FAUCET
    initial_mint: Optional[int] = MAX_FAUCET


@dataclass
class MarketMakerArgs:
    market_kappa: float
    market_order_arrival_rate: int
    order_kappa: float
    order_levels: int
    order_spacing: float
    order_size: int
    order_clipping: int
    inventory_upper_boundary: int
    inventory_lower_boundary: int
    fee_amount: float
    commitment_amount: int
    initial_mint: Optional[int] = MAX_FAUCET


@dataclass
class AuctionTraderArgs:
    initial_volume: float
    initial_mint: Optional[int] = MAX_FAUCET


@dataclass
class RandomTraderArgs:
    order_intensity: List[int]
    order_volume: List[float]
    step_bias: List[float]
    initial_mint: Optional[int] = MAX_FAUCET


@dataclass
class SensitiveTraderArgs:
    scale: List[int]
    max_order_size: List[float]
    initial_mint: Optional[int] = MAX_FAUCET


@dataclass
class SimulationArgs:
    n_steps: int
    granularity: Granularity
    coinbase_code: str
    start_date: Optional[str] = "2022-08-01 00:00:00"
    randomise_history: Optional[bool] = False



import os
import ast

import itertools
import argparse

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import vega_sim.proto.vega as vega_protos

from typing import Optional, List
from collections import defaultdict
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.markers import MarkerStyle
from matplotlib.gridspec import GridSpec, SubplotSpec, GridSpecFromSubplotSpec

from vega_sim.scenario.fuzzed_markets.agents import (
    FuzzingAgent,
    FuzzyLiquidityProvider,
    RiskyMarketOrderTrader,
    RiskySimpleLiquidityProvider,
)

from vega_sim.proto.vega import markets

import numpy as np


TRADING_MODE_COLORS = {
    0: (200 / 255, 200 / 255, 200 / 255),
    1: (204 / 255, 255 / 255, 15 / 2553),
    2: (255 / 255, 204 / 255, 153 / 255),
    3: (153 / 255, 204 / 255, 244 / 255),
    4: (255 / 255, 133 / 255, 133 / 255),
    5: (255 / 255, 204 / 255, 255 / 255),
    6: (255 / 255, 0 / 255, 0 / 255),
}


"""
Thoughts for plots
    - Total LP Margin positions
    - Market State
"""

from vega_sim.tools.scenario_output import (
    load_accounts_df,
    load_market_data_df,
    load_order_book_df,
    load_trades_df,
    load_fuzzing_df,
    load_agents_df,
    load_market_chain,
    load_resource_df,
    load_assets_df,
)


def _set_xticks(ax, index, freq: int = 60):
    xaxis_ticks = index[0:-1:freq]
    labels = [dt.strftime("%H:%M:%S") for dt in xaxis_ticks.to_pydatetime()]
    ax.set_xticks(
        xaxis_ticks,
        labels=labels,
        rotation=45,
    )


def _order_book_df_to_heatmap_df(order_book_df: pd.DataFrame) -> pd.DataFrame:
    book_data_df = order_book_df.reset_index()[["time", "price", "volume"]].copy()

    min_pr = order_book_df.price.min()
    max_pr = order_book_df.price.max()
    int_range = pd.interval_range(start=min_pr, end=max_pr, freq=0.1)
    book_data_df["interval"] = pd.cut(book_data_df.price, bins=int_range)
    book_data_df = (
        book_data_df.groupby(["time", "interval"])[["volume"]].sum().sort_index()
    )
    idx = pd.MultiIndex.from_product(
        [
            book_data_df.index.get_level_values(0).unique(),
            book_data_df.index.get_level_values(1).unique(),
        ],
        names=["time", "interval"],
    )

    return (
        book_data_df.reindex(idx, fill_value=0)
        .sort_index(ascending=[True, False])
        .unstack(level=-1)
        .transpose()
        .sort_index(ascending=False)
        .values
    )


def _series_df_to_single_series(
    df: pd.DataFrame, market_series: List[str]
) -> pd.DataFrame:
    df_clone = df.copy().reset_index()
    df_clone["series_ix"] = df_clone.market_id.apply(lambda x: market_series.index(x))
    selector = df_clone.loc[df_clone.groupby("time").series_ix.idxmax()][
        ["time", "market_id"]
    ]

    return df_clone.merge(selector, on=["time", "market_id"]).set_index("time")


def plot_trading_summary(
    ax: Axes,
    trade_df: pd.DataFrame,
    order_book_df: pd.DataFrame,
    price_df: pd.DataFrame,
) -> None:
    if any(df.empty for df in (trade_df, order_book_df, price_df)):
        return
    buys = trade_df[trade_df.aggressor == 1]
    sells = trade_df[trade_df.aggressor == 2]

    heatmap_df = _order_book_df_to_heatmap_df(order_book_df=order_book_df)

    ax.set_title("Trading Summary")

    ax.imshow(
        heatmap_df,
        interpolation="none",
        aspect="auto",
        extent=(
            int(price_df.index[0].value / 1e9),
            int(price_df.index[-1].value / 1e9),
            order_book_df.price.min(),
            order_book_df.price.max(),
        ),
    )
    x_ticks = [t.value / 1e9 for t in price_df.index]
    ax.plot(x_ticks, price_df["price"])
    ax.scatter(
        buys["seen_at"], buys["price"], marker=MarkerStyle("x").scaled(0.5), c="red"
    )
    ax.scatter(
        sells["seen_at"], sells["price"], marker=MarkerStyle("s").scaled(0.5), c="blue"
    )
    ax.set_xticks(
        x_ticks[0:-1:60],
        labels=[
            dt.strftime("%H:%M:%S") for dt in price_df.index.to_pydatetime()[0:-1:60]
        ],
        rotation=45,
    )


def plot_total_traded_volume(ax: Axes, trades_df: pd.DataFrame) -> None:
    if trades_df.empty:
        return
    ax.set_title("Traded Volume")
    traded = trades_df.groupby(level=0)["size"].sum().cumsum()

    _set_xticks(ax, traded.index)

    ax.plot(traded)


def plot_open_interest(ax: Axes, market_data_df: pd.DataFrame):
    ax.set_title("Open Interest")

    _set_xticks(ax, market_data_df.index)
    ax.plot(market_data_df["open_interest"])


def plot_open_notional(
    ax: Axes,
    market_data_df: pd.DataFrame,
    price_df: pd.DataFrame,
) -> None:
    if any(df.empty for df in (market_data_df, price_df)):
        return
    ax.set_title("Open Notional $")

    _set_xticks(ax, market_data_df.index)
    ax.plot(market_data_df["open_interest"] * price_df["price"])


def plot_spread(ax: Axes, order_book_df: pd.DataFrame) -> None:
    if order_book_df.empty:
        return
    ax.set_title("10s Rolling Avg Spread")

    spread_df = (
        order_book_df[order_book_df.level == 0]
        .groupby(level=0)["price"]
        .diff()
        .dropna()
        .rolling(10)
        .mean()
    )

    _set_xticks(ax, spread_df.index)
    ax.plot(spread_df)


def plot_margin_totals(ax: Axes, accounts_df: pd.DataFrame) -> None:
    if accounts_df.empty:
        return
    ax.set_title("Total Margin Balance")
    grouped_df = (
        accounts_df[accounts_df.type == vega_protos.vega.ACCOUNT_TYPE_MARGIN]
        .groupby(level=0)[["balance"]]
        .sum()
        .iloc[1:]
    )
    _set_xticks(ax, grouped_df.index)
    ax.plot(grouped_df)


def plot_target_stake(ax: Axes, market_data_df: pd.DataFrame) -> None:
    if market_data_df.empty:
        return
    ax.set_title("Target Stake")

    _set_xticks(ax, market_data_df.index)
    ax.plot(market_data_df["target_stake"])


def plot_run_outputs(run_name: Optional[str] = None) -> list[Figure]:
    order_df = load_order_book_df(run_name=run_name)
    trades_df = load_trades_df(run_name=run_name)
    data_df = load_market_data_df(run_name=run_name)
    accounts_df = load_accounts_df(run_name=run_name)

    market_chains = load_market_chain(run_name=run_name)
    if not market_chains:
        market_chains = {
            market_id: [market_id] for market_id in data_df["market_id"].unique()
        }

    figs = {}

    for market_id, market_children in market_chains.items():
        market_order_df = _series_df_to_single_series(
            order_df[order_df["market_id"].isin(market_children)],
            market_series=market_children,
        )
        market_trades_df = _series_df_to_single_series(
            trades_df[trades_df["market_id"].isin(market_children)],
            market_series=market_children,
        )
        market_data_df = _series_df_to_single_series(
            data_df[data_df["market_id"].isin(market_children)],
            market_series=market_children,
        )
        market_accounts_df = _series_df_to_single_series(
            accounts_df[accounts_df["market_id"].isin(market_children)],
            market_series=market_children,
        )

        market_mid_df = (
            market_order_df[market_order_df.level == 0].groupby("time")[["price"]].sum()
            / 2
        )

        fig = plt.figure(layout="tight", figsize=(8, 10))

        ax = plt.subplot(411)
        ax2 = plt.subplot(423)
        ax3 = plt.subplot(424)
        ax4 = plt.subplot(425)
        ax5 = plt.subplot(426)
        ax6 = plt.subplot(427)
        ax7 = plt.subplot(428)

        plot_trading_summary(ax, market_trades_df, market_order_df, market_mid_df)
        plot_total_traded_volume(ax2, market_trades_df)
        plot_spread(ax3, order_book_df=market_order_df)
        plot_open_interest(ax4, market_data_df)
        plot_open_notional(ax5, market_data_df=market_data_df, price_df=market_mid_df)
        plot_margin_totals(ax6, accounts_df=market_accounts_df)
        plot_target_stake(ax7, market_data_df=market_data_df)

        figs[market_id] = fig

    return figs


def plot_trading_mode(
    fig: Figure, data_df: pd.DataFrame, ss: Optional[SubplotSpec] = None
):
    """Plots the proportion of time spent in different trading modes of a market.

    Args:
        fig (Figure):
            The Figure object to which the subplots will be added.
        data_df (pd.DataFrame):
            The DataFrame containing market data, including the trading mode of the
            market at each time step.
        gs (Optional[SubplotSpec]):
            An optional SubplotSpec object defining the placement of the subplots. If
            not specified, a default GridSpec with two rows and one column will be used.

    The function adds two subplots to the Figure object fig. The top subplot shows a
    step plot with a fill between each step representing the trading mode at the current
    time step. The bottom subplot shows a stacked area plot of the same information,
    with each area representing the proportion of time spent in a particular trading
    mode. The legend in the bottom subplot shows the name of each trading mode.

    TRADING_MODE_COLORS is a dictionary that maps each trading mode to a color used in
    the plots.
    """

    if ss is None:
        gs = GridSpec(
            nrows=2,
            ncols=1,
            height_ratios=[1, 5],
            hspace=0.1,
        )
    else:
        gs = GridSpecFromSubplotSpec(
            subplot_spec=ss,
            nrows=2,
            ncols=1,
            height_ratios=[1, 5],
            hspace=0.1,
        )
    ax0 = fig.add_subplot(gs[0, 0])
    ax1 = fig.add_subplot(gs[1, 0])

    ax0.set_title(
        "Trading Mode Analysis", loc="left", fontsize=12, color=(0.3, 0.3, 0.3)
    )
    names = []
    for name, value in markets.Market.TradingMode.items():
        names.append(name)
        series = (data_df["market_trading_mode"] == value).astype(int)
        ax0.fill_between(
            series.index,
            series,
            step="post",
            alpha=1,
            color=TRADING_MODE_COLORS[value],
            linewidth=0,
        )

        data_df = data_df.merge(
            (data_df["market_trading_mode"] == value).cumsum().rename(name),
            left_index=True,
            right_index=True,
        )
    data_df = data_df[names].divide(data_df[names].sum(axis=1), axis=0)

    ax1.stackplot(
        data_df.index,
        *[data_df[name].values for name in names],
        colors=TRADING_MODE_COLORS.values(),
    )

    ax0.get_xaxis().set_visible(False)
    ax0.get_yaxis().set_visible(False)

    ax1.set_ylabel("PROPORTION OF TIME IN MODE")
    ax1.legend(labels=names, loc="lower right")


def plot_price_comparison(
    fig: Figure,
    data_df: pd.DataFrame,
    fuzzing_df: pd.DataFrame,
    ss: Optional[SubplotSpec],
):
    """Plots the external price and mark price along with their respective volatilities.

    Args:
        fig (Figure):
            Figure object to plot the data on.
        fuzzing_df (pd.DataFrame):
            DataFrame containing fuzzing data.
        data_df (pd.DataFrame):
            DataFrame containing market data.
        ss (Optional[SubplotSpec]):
            SubplotSpec object representing the position of the subplot on the figure.
    """
    if ss is None:
        gs = GridSpec(
            nrows=1,
            ncols=1,
        )
    else:
        gs = GridSpecFromSubplotSpec(
            subplot_spec=ss,
            nrows=1,
            ncols=1,
        )
    ax0 = fig.add_subplot(gs[0, 0])

    ax0.set_title("Price Analysis", loc="left", fontsize=12, color=(0.3, 0.3, 0.3))

    external_price_series = fuzzing_df["external_price"].replace(0, np.nan)
    mark_price_series = data_df["mark_price"].replace(0, np.nan)

    ax0.plot(mark_price_series)
    ax0.set_ylim(ax0.get_ylim())
    ax0.plot(external_price_series, linewidth=0.8, alpha=0.8)

    ep_volatility = external_price_series.var() / external_price_series.size
    mp_volatility = mark_price_series.var() / mark_price_series.size

    ax0.text(
        x=0.1,
        y=0.1,
        s=(
            f"external-price volatility = {round(ep_volatility, 1)}\nmark-price"
            f" volatility = {round(mp_volatility, 1)}"
        ),
        fontsize=8,
        bbox=dict(facecolor="white", alpha=1),
        transform=ax0.transAxes,
    )

    ax0.set_ylabel("PRICE")
    ax0.legend(labels=["external price", "mark price"])


def plot_risky_close_outs(
    fig: Figure,
    accounts_df: pd.DataFrame,
    fuzzing_df: pd.DataFrame,
    ss: Optional[SubplotSpec] = None,
):
    """Plots the number of close outs of risky traders and risky liquidity providers.

    Args:
        fig (matplotlib.figure.Figure):
            The figure object to plot onto.
        accounts_df (pandas.DataFrame):
            A dataframe containing the accounts data.
        fuzzing_df (pandas.DataFrame):
            A dataframe containing the fuzzing data.
        ss (Optional[matplotlib.gridspec.SubplotSpec]):
            A subplot specification for the plot. Default is None.
    """
    if ss is None:
        gs = GridSpec(
            nrows=2,
            ncols=1,
            height_ratios=[1, 1],
            hspace=0.15,
        )
    else:
        gs = GridSpecFromSubplotSpec(
            subplot_spec=ss,
            nrows=2,
            ncols=1,
            height_ratios=[1, 1],
            hspace=0.15,
        )

    ax0 = fig.add_subplot(gs[0, 0])
    ax1 = fig.add_subplot(gs[1, 0])

    ax0.set_title("Close Out Analysis", loc="left", fontsize=12, color=(0.3, 0.3, 0.3))

    insurance_pool_ds = accounts_df["balance"][
        (accounts_df["party_id"] == "network") & (accounts_df["type"] == 1)
    ]

    trader_close_outs_ds = fuzzing_df["trader_close_outs"]
    liquidity_provider_close_outs_ds = fuzzing_df["liquidity_provider_close_outs"]

    ax0r = ax0.twinx()

    ln0 = ax0r.plot(
        insurance_pool_ds.index,
        insurance_pool_ds.values,
        "b.-",
        markersize=1,
        label="insurance pool",
    )
    ln1 = ax0.plot(
        trader_close_outs_ds.index,
        trader_close_outs_ds.values,
        "r.-",
        markersize=1,
        label="risky trader close outs",
    )

    lns = ln0 + ln1
    ax0.legend(handles=lns, labels=[ln.get_label() for ln in lns], loc="upper left")

    ax0.set_ylabel("NB CLOSE OUTS")
    ax0r.set_ylabel("INSURANCE POOL", position="right")
    ax0r.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))

    ax0.get_xaxis().set_visible(False)

    ax1r = ax1.twinx()

    ln3 = ax1r.plot(
        insurance_pool_ds.index,
        insurance_pool_ds.values,
        "b.-",
        markersize=1,
        label="insurance pool",
    )
    ln4 = ax1.plot(
        liquidity_provider_close_outs_ds.index,
        liquidity_provider_close_outs_ds.values,
        "r.-",
        markersize=1,
        label="risky LP close outs",
    )

    lns = ln3 + ln4
    ax1.legend(handles=lns, labels=[ln.get_label() for ln in lns], loc="upper left")

    ax1.set_ylabel("NB CLOSE OUTS")
    ax1r.set_ylabel("INSURANCE POOL", position="right")
    ax1r.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))


def fuzz_plots(run_name: Optional[str] = None) -> Figure:
    data_df = load_market_data_df(run_name=run_name)
    accounts_df = load_accounts_df(run_name=run_name)
    fuzzing_df = load_fuzzing_df(run_name=run_name)

    market_chains = load_market_chain(run_name=run_name)
    if not market_chains:
        market_chains = {
            market_id: [market_id] for market_id in data_df["market_id"].unique()
        }

    figs = {}
    for market_id, market_children in market_chains.items():
        market_data_df = _series_df_to_single_series(
            data_df[data_df["market_id"].isin(market_children)],
            market_series=market_children,
        )
        market_accounts_df = _series_df_to_single_series(
            accounts_df[accounts_df["market_id"].isin(market_children)],
            market_series=market_children,
        )

        market_fuzzing_df = _series_df_to_single_series(
            fuzzing_df[fuzzing_df["market_id"].isin(market_children)],
            market_series=market_children,
        )

        fig = plt.figure(figsize=[8, 10])
        fig.suptitle(
            f"Fuzz Testing Plots",
            fontsize=18,
            fontweight="bold",
            color=(0.2, 0.2, 0.2),
        )
        fig.tight_layout()

        plt.rcParams.update({"font.size": 8})

        gs = GridSpec(nrows=3, ncols=1, height_ratios=[2, 2, 3], hspace=0.3)

        plot_trading_mode(fig, ss=gs[0, 0], data_df=market_data_df)
        plot_price_comparison(
            fig,
            ss=gs[1, 0],
            data_df=market_data_df,
            fuzzing_df=market_fuzzing_df,
        )
        plot_risky_close_outs(
            fig,
            ss=gs[2, 0],
            accounts_df=market_accounts_df,
            fuzzing_df=market_fuzzing_df,
        )

        figs[market_id] = fig

    return figs


def account_plots(run_name: Optional[str] = None, agent_types: Optional[list] = None):
    accounts_df = load_accounts_df(run_name=run_name)
    agents_df = load_agents_df(run_name=run_name)

    fig = plt.figure(figsize=[8, 10])
    fig.suptitle(
        f"Agent Account Plots",
        fontsize=18,
        fontweight="bold",
        color=(0.2, 0.2, 0.2),
    )
    fig.tight_layout()

    plt.rcParams.update({"font.size": 8})

    agent_types = [
        FuzzingAgent,
        FuzzyLiquidityProvider,
        RiskyMarketOrderTrader,
        RiskySimpleLiquidityProvider,
    ]

    gs = GridSpec(nrows=len(agent_types), ncols=1, hspace=0.5)

    axs: list[plt.Axes] = []

    for i, agent_type in enumerate(agent_types):
        axs.append(fig.add_subplot(gs[i, 0]))

        axs[i].set_title(
            f"Total Account Balance: {agent_type.__name__}",
            loc="left",
            fontsize=12,
            color=(0.3, 0.3, 0.3),
        )

        agent_keys = agents_df["agent_key"][
            agents_df["agent_type"] == agent_type.__name__
        ].to_list()
        for key in agent_keys:
            totals = (
                accounts_df[accounts_df["party_id"] == key]["balance"]
                .groupby(level=0)
                .sum()
            )
            axs[i].plot(totals)
        axs[i].autoscale(enable=True, axis="y")

    return fig


def plot_price_monitoring(run_name: Optional[str] = None):
    data_df = load_market_data_df(run_name=run_name)
    market_chains = load_market_chain(run_name=run_name)
    if not market_chains:
        market_chains = {
            market_id: [market_id] for market_id in data_df["market_id"].unique()
        }

    figs = {}
    for market_id, market_children in market_chains.items():
        market_data_df = _series_df_to_single_series(
            data_df[data_df["market_id"].isin(market_children)],
            market_series=market_children,
        )

        # Extract the tightest valid prices from the price monitoring valid_prices
        valid_prices = defaultdict(lambda: [])
        for index in market_data_df.index:
            all_bounds = list(
                itertools.chain(
                    *[
                        ast.literal_eval(
                            bounds[0] if isinstance(bounds, np.ndarray) else bounds
                        )
                        for bounds in market_data_df.loc[index][
                            ["price_monitoring_bounds"]
                        ].values
                    ]
                )
            )
            valid_prices["datetime"].append(index)
            valid_prices["min_valid_price"].append(np.nan)
            valid_prices["max_valid_price"].append(np.nan)

            for _, individual_bound in enumerate(all_bounds):
                valid_prices["min_valid_price"][-1] = (
                    individual_bound[0]
                    if valid_prices["min_valid_price"][-1] is np.nan
                    else max(individual_bound[0], valid_prices["min_valid_price"][-1])
                )
                valid_prices["max_valid_price"][-1] = (
                    individual_bound[1]
                    if valid_prices["max_valid_price"][-1] is np.nan
                    else min(individual_bound[1], valid_prices["max_valid_price"][-1])
                )

        fig = plt.figure(figsize=[10, 7])
        fig.suptitle(
            f"Price Monitoring Analysis",
            fontsize=18,
            fontweight="bold",
            color=(0.2, 0.2, 0.2),
        )
        fig.tight_layout()
        plt.rcParams.update({"font.size": 8})

        gs = GridSpec(nrows=1, ncols=1, hspace=0.1)
        ax = fig.add_subplot(
            gs[0, 0],
        )
        twinax = ax.twinx()
        twinax.set_ylim(0, 1)

        # Plot period where auctions triggered (but not extended)
        series = (market_data_df["trigger"] == 3).astype(int) & (
            market_data_df["extension_trigger"] != 3
        ).astype(int)
        twinax.fill_between(
            series.index,
            series,
            step="post",
            alpha=0.1,
            color="r",
            linewidth=0,
            label="auction",
        )
        # Plot periods where auctions extended
        series = (market_data_df["extension_trigger"] == 3).astype(int)
        twinax.fill_between(
            series.index,
            series,
            step="post",
            alpha=0.1,
            color="orange",
            linewidth=0,
            label="extension",
        )

        ax.plot(
            valid_prices["datetime"],
            valid_prices["min_valid_price"],
            "r-",
            linewidth=0.5,
            alpha=0.4,
            label="valid price bounds",
        )
        ax.plot(
            valid_prices["datetime"],
            valid_prices["max_valid_price"],
            "r-",
            linewidth=0.5,
            alpha=0.4,
            label="_nolegend",
        )

        ax.plot(
            market_data_df.index,
            market_data_df["mark_price"].replace(0, np.nan),
            "b-",
            alpha=1.0,
            label="mark price",
        )
        ax.plot(
            market_data_df.index,
            market_data_df["mid_price"].replace(0, np.nan),
            "b-",
            alpha=0.4,
            label="mid price",
        )
        ax.plot(
            market_data_df.index,
            market_data_df["indicative_price"].replace(0, np.nan),
            "g-",
            alpha=1.0,
            label="indicative price",
        )

        ax.legend(loc="upper left")
        twinax.legend(loc="lower right")

        ax.set_xlabel("datetime")
        ax.set_ylabel("price")

        figs[market_id] = fig

    return figs


def resource_monitoring_plot(run_name: Optional[str] = None):
    resource_df = load_resource_df(
        run_name=run_name,
    )

    fig = plt.figure(figsize=[11.69, 8.27])
    fig.suptitle(
        f"Resource Monitoring",
        fontsize=18,
        fontweight="bold",
        color=(0.2, 0.2, 0.2),
    )
    fig.tight_layout()
    plt.rcParams.update({"font.size": 8})
    plt.rcParams.update({"axes.formatter.useoffset": False})

    gs = GridSpec(nrows=3, ncols=2, hspace=0.5, wspace=0.3)

    ax0 = fig.add_subplot(
        gs[0, :],
    )
    ax0.set_title("CPU Utilization")
    ax0.plot(resource_df.index, resource_df.vega_cpu_per, "r", label="vega")
    ax0.plot(resource_df.index, resource_df.datanode_cpu_per, "b", label="data-node")
    ax0.legend()
    ax0.set_xlabel("'vega' datetime")
    ax0.set_ylabel("CPU [%]")

    ax1 = fig.add_subplot(gs[1, 0], sharex=ax0)
    ax1.set_title("Memory - RSS (vega)")
    ax1.plot(resource_df.index, resource_df.vega_mem_rss / 1e9, "r", label="vega")
    ax1.legend()
    ax1.set_xlabel("'vega' datetime")
    ax1.set_ylabel("RSS [GB]")

    ax2 = fig.add_subplot(gs[1, 1], sharex=ax0)
    ax2.set_title("Memory - RSS (data-node)")
    ax2.plot(
        resource_df.index, resource_df.datanode_mem_rss / 1e9, "b", label="data-node"
    )
    ax2.legend()
    ax2.set_xlabel("'vega' datetime")
    ax2.set_ylabel("RSS [GB]")

    ax3 = fig.add_subplot(gs[2, 0], sharex=ax0)
    ax3.set_title("Memory - VMS (vega)")
    ax3.plot(resource_df.index, resource_df.vega_mem_vms / 1e9, "r", label="vega")
    ax3.legend()
    ax3.set_xlabel("'vega' datetime")
    ax3.set_ylabel("VMS [GB]")

    ax4 = fig.add_subplot(gs[2, 1], sharex=ax0)
    ax4.set_title("Memory - VMS (data-node)")
    ax4.plot(
        resource_df.index, resource_df.datanode_mem_vms / 1e9, "b", label="data-node"
    )
    ax4.legend()
    ax4.set_xlabel("'vega' datetime")
    ax4.set_ylabel("VMS [GB]")

    return fig


def reward_plots(run_name: Optional[str] = None):
    accounts_df = load_accounts_df(run_name=run_name)
    assets_df = load_assets_df(run_name=run_name)
    agents_df = load_agents_df(run_name=run_name).set_index("agent_key")

    joined_df = accounts_df.join(agents_df, on="party_id").join(assets_df, on="asset")

    datetime = joined_df.index.unique()

    fig = plt.figure(figsize=[11.69, 8.27])
    fig.suptitle(
        f"Agent Rewards Plots",
        fontsize=18,
        fontweight="bold",
        color=(0.2, 0.2, 0.2),
    )
    fig.tight_layout()

    plt.rcParams.update({"font.size": 8})

    gs = GridSpec(nrows=2, ncols=2, hspace=0.4)

    axs: list[plt.Axes] = []

    plots = [
        (
            0,
            0,
            vega_protos.vega.DispatchMetric.Name(
                vega_protos.vega.DISPATCH_METRIC_MAKER_FEES_PAID
            ),
        ),
        (
            0,
            1,
            vega_protos.vega.DispatchMetric.Name(
                vega_protos.vega.DISPATCH_METRIC_MAKER_FEES_RECEIVED
            ),
        ),
        (
            1,
            0,
            vega_protos.vega.DispatchMetric.Name(
                vega_protos.vega.DISPATCH_METRIC_LP_FEES_RECEIVED
            ),
        ),
        (
            1,
            1,
            vega_protos.vega.DispatchMetric.Name(
                vega_protos.vega.DISPATCH_METRIC_MARKET_VALUE
            ),
        ),
    ]

    for plot in plots:
        axs.append(fig.add_subplot(gs[plot[0], plot[1]]))
        axs[-1].set_title(f"Asset: {plot[2]}")
        accounts_for_asset = joined_df[joined_df.symbol == str(plot[2])]
        grouped = accounts_for_asset.groupby(["agent_type", "time"])["balance"].sum()
        for index in grouped.index.get_level_values(0).unique():
            if index == "RewardFunder":
                continue
            series = grouped.loc[index]
            series = series.reindex(datetime, fill_value=0)
            axs[-1].plot(series.index, series.values, label=index)
        plt.xticks(rotation=45)
        axs[-1].legend()

    return fig


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--fuzzing", action="store_true")
    parser.add_argument("--trading", action="store_true")
    parser.add_argument("--monitoring", action="store_true")
    parser.add_argument("--accounts", action="store_true")
    parser.add_argument("--rewards", action="store_true")
    parser.add_argument("--resources", action="store_true")
    parser.add_argument("--all", action="store_true")

    parser.add_argument("--show", action="store_true")
    parser.add_argument("--save", action="store_true")
    args = parser.parse_args()

    dir = "test_plots"
    if not os.path.exists(dir):
        os.mkdir(dir)

    if args.fuzzing or args.all:
        figs = fuzz_plots()
        for i, fig in enumerate(figs.values()):
            if args.save:
                fig.savefig(f"{dir}/fuzz-{i}.jpg")

    if args.trading or args.all:
        figs = plot_run_outputs()
        for i, fig in enumerate(figs.values()):
            if args.save:
                fig.savefig(f"{dir}/trading-{i}.jpg")

    if args.monitoring or args.all:
        figs = plot_price_monitoring()
        for i, fig in enumerate(figs.values()):
            if args.save:
                fig.savefig(f"{dir}/monitoring-{i}.jpg")

    if args.accounts or args.all:
        print("accounting")
        fig = account_plots()
        if args.save:
            fig.savefig(f"{dir}/accounts.jpg")

    if args.rewards or args.all:
        fig = reward_plots()
        if args.save:
            fig.savefig(f"{dir}/rewards.jpg")

    if args.resources or args.all:
        fig = resource_monitoring_plot()
        if args.save:
            fig.savefig(f"{dir}/resource_monitoring.jpg")

    if args.show:
        plt.show()



import logging
from collections import namedtuple
from dataclasses import dataclass
from typing import (
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Callable,
    TypeVar,
    Union,
)
import datetime
import vega_sim.api.data_raw as data_raw
import vega_sim.grpc.client as vac
import vega_sim.proto.data_node.api.v2 as data_node_protos_v2
import vega_sim.proto.vega as vega_protos
import vega_sim.proto.vega.events.v1.events_pb2 as events_protos
from vega_sim.api.helpers import num_from_padded_int


class MissingAssetError(Exception):
    pass


class MissingMarketError(Exception):
    pass


logger = logging.Logger(__name__)

T = TypeVar("T")
S = TypeVar("S")

PartyMarketAccount = namedtuple("PartyMarketAccount", ["general", "margin", "bond"])
AccountData = namedtuple(
    "AccountData", ["owner", "balance", "asset", "market_id", "type"]
)
RiskFactor = namedtuple("RiskFactors", ["market_id", "short", "long"])
OrderBook = namedtuple("OrderBook", ["bids", "asks"])
PriceLevel = namedtuple("PriceLevel", ["price", "number_of_orders", "volume"])
Order = namedtuple(
    "Order",
    [
        "price",
        "size",
        "id",
        "reference",
        "side",
        "status",
        "remaining",
        "time_in_force",
        "order_type",
        "created_at",
        "expires_at",
        "party_id",
        "market_id",
        "updated_at",
        "version",
    ],
)
Position = namedtuple(
    "Position",
    [
        "party_id",
        "market_id",
        "open_volume",
        "realised_pnl",
        "unrealised_pnl",
        "average_entry_price",
        "updated_at",
        "loss_socialisation_amount",
    ],
)
Transfer = namedtuple(
    "Transfer",
    [
        "id",
        "party_from",
        "from_account_type",
        "party_to",
        "to_account_type",
        "asset",
        "amount",
        "reference",
        "status",
        "timestamp",
        "reason",
        "one_off",
        "recurring",
    ],
)

MarginLevels = namedtuple(
    "MarginLevels",
    [
        "maintenance_margin",
        "search_level",
        "initial_margin",
        "collateral_release_level",
        "party_id",
        "market_id",
        "asset",
        "timestamp",
    ],
)


@dataclass
class LiquidationPrice:
    open_volume_only: float
    including_buy_orders: float
    including_sell_orders: float


@dataclass
class MarginEstimate:
    best_case: MarginLevels
    worst_case: MarginLevels


@dataclass
class LiquidationEstimate:
    best_case: LiquidationPrice
    worst_case: LiquidationPrice


@dataclass
class DecimalSpec:
    position_decimals: Optional[int] = None
    price_decimals: Optional[int] = None
    asset_decimals: Optional[int] = None


@dataclass
class LedgerEntry:
    from_account: vega_protos.vega.AccountDetails
    to_account: vega_protos.vega.AccountDetails
    amount: str
    transfer_type: vega_protos.vega.TransferType
    timestamp: int


@dataclass
class AggregatedLedgerEntry:
    timestamp: int
    quantity: float
    transfer_type: vega_protos.vega.TransferType
    asset_id: str
    from_account_type: vega_protos.vega.AccountType
    to_account_type: vega_protos.vega.AccountType
    from_account_party_id: str
    to_account_party_id: str
    from_account_market_id: str
    to_account_market_id: str


@dataclass
class MarketDepth:
    buys: List[PriceLevel]
    sells: List[PriceLevel]


@dataclass
class OrdersBySide:
    bids: List[Order]
    asks: List[Order]


@dataclass
class Fee:
    maker_fee: float
    infrastructure_fee: float
    liquidity_fee: float


@dataclass(frozen=True)
class Trade:
    id: str
    market_id: str
    price: float
    size: float
    buyer: str
    seller: str
    aggressor: vega_protos.vega.Side
    buy_order: str
    sell_order: str
    timestamp: int
    trade_type: vega_protos.vega.Trade.Type
    buyer_fee: Fee
    seller_fee: Fee
    buyer_auction_batch: int
    seller_auction_batch: int


@dataclass(frozen=True)
class MarketData:
    mark_price: float
    best_bid_price: float
    best_bid_volume: float
    best_offer_price: float
    best_offer_volume: float
    best_static_bid_price: float
    best_static_bid_volume: float
    best_static_offer_price: float
    best_static_offer_volume: float
    mid_price: float
    static_mid_price: float
    market_id: str
    timestamp: str
    open_interest: float
    auction_end: str
    auction_start: str
    indicative_price: float
    indicative_volume: float
    market_trading_mode: str
    trigger: str
    extension_trigger: str
    target_stake: float
    supplied_stake: float
    market_value_proxy: float
    price_monitoring_bounds: list
    liquidity_provider_fee_share: list
    market_state: str
    next_mark_to_market: float
    last_traded_price: float


@dataclass(frozen=True)
class PriceMonitoringBounds:
    min_valid_price: float
    max_valid_price: float
    trigger: str
    reference_price: float


@dataclass(frozen=True)
class LiquidityProviderFeeShare:
    party: str
    equity_like_share: float
    average_entry_valuation: float
    average_score: float


def _ledger_entry_from_proto(
    ledger_entry,
    asset_decimals: int,
) -> LedgerEntry:
    return LedgerEntry(
        from_account=ledger_entry.from_account,
        to_account=ledger_entry.to_account,
        amount=num_from_padded_int(ledger_entry.amount, asset_decimals),
        transfer_type=ledger_entry.type,
        timestamp=ledger_entry.timestamp,
    )


def _aggregated_ledger_entry_from_proto(
    ledger_entry: data_node_protos_v2.trading_data.AggregatedLedgerEntry,
    decimal_spec: DecimalSpec,
) -> AggregatedLedgerEntry:
    return AggregatedLedgerEntry(
        timestamp=ledger_entry.timestamp,
        quantity=num_from_padded_int(
            ledger_entry.quantity, decimal_spec.asset_decimals
        ),
        transfer_type=ledger_entry.transfer_type,
        asset_id=ledger_entry.asset_id,
        from_account_type=ledger_entry.from_account_type,
        to_account_type=ledger_entry.to_account_type,
        from_account_party_id=ledger_entry.from_account_party_id,
        to_account_party_id=ledger_entry.to_account_party_id,
        from_account_market_id=ledger_entry.from_account_market_id,
        to_account_market_id=ledger_entry.to_account_market_id,
    )


def _trade_from_proto(
    trade: vega_protos.vega.Trade,
    decimal_spec: DecimalSpec,
) -> Trade:
    return Trade(
        id=trade.id,
        market_id=trade.market_id,
        price=num_from_padded_int(trade.price, decimal_spec.price_decimals),
        size=num_from_padded_int(trade.size, decimal_spec.position_decimals),
        buyer=trade.buyer,
        seller=trade.seller,
        aggressor=trade.aggressor,
        buy_order=trade.buy_order,
        sell_order=trade.sell_order,
        timestamp=trade.timestamp,
        trade_type=trade.type,
        buyer_fee=Fee(
            maker_fee=num_from_padded_int(
                trade.buyer_fee.maker_fee, decimal_spec.asset_decimals
            ),
            infrastructure_fee=num_from_padded_int(
                trade.buyer_fee.infrastructure_fee, decimal_spec.asset_decimals
            ),
            liquidity_fee=num_from_padded_int(
                trade.buyer_fee.liquidity_fee, decimal_spec.asset_decimals
            ),
        ),
        seller_fee=Fee(
            maker_fee=num_from_padded_int(
                trade.seller_fee.maker_fee, decimal_spec.asset_decimals
            ),
            infrastructure_fee=num_from_padded_int(
                trade.seller_fee.infrastructure_fee, decimal_spec.asset_decimals
            ),
            liquidity_fee=num_from_padded_int(
                trade.seller_fee.liquidity_fee, decimal_spec.asset_decimals
            ),
        ),
        buyer_auction_batch=trade.buyer_auction_batch,
        seller_auction_batch=trade.seller_auction_batch,
    )


def _margin_level_from_proto(
    margin_level: vega_protos.vega.MarginLevels, decimal_spec: DecimalSpec
) -> MarginLevels:
    return MarginLevels(
        maintenance_margin=num_from_padded_int(
            margin_level.maintenance_margin, decimal_spec.asset_decimals
        ),
        search_level=num_from_padded_int(
            margin_level.search_level, decimal_spec.asset_decimals
        ),
        initial_margin=num_from_padded_int(
            margin_level.initial_margin, decimal_spec.asset_decimals
        ),
        collateral_release_level=num_from_padded_int(
            margin_level.collateral_release_level, decimal_spec.asset_decimals
        ),
        party_id=margin_level.party_id,
        market_id=margin_level.market_id,
        asset=margin_level.asset,
        timestamp=margin_level.timestamp,
    )


def _order_from_proto(
    order: vega_protos.vega.Order,
    decimal_spec: DecimalSpec,
) -> Order:
    return Order(
        id=order.id,
        price=num_from_padded_int(order.price, decimal_spec.price_decimals),
        size=num_from_padded_int(order.size, decimal_spec.position_decimals),
        reference=order.reference,
        side=order.side,
        status=order.status,
        remaining=num_from_padded_int(order.remaining, decimal_spec.position_decimals),
        time_in_force=order.time_in_force,
        order_type=order.type,
        created_at=order.created_at,
        expires_at=order.expires_at,
        party_id=order.party_id,
        updated_at=order.updated_at,
        version=order.version,
        market_id=order.market_id,
    )


def _position_from_proto(
    position: vega_protos.vega.Position,
    decimal_spec: DecimalSpec,
) -> Position:
    return Position(
        party_id=position.party_id,
        market_id=position.market_id,
        open_volume=num_from_padded_int(
            position.open_volume, decimal_spec.position_decimals
        ),
        realised_pnl=num_from_padded_int(
            position.realised_pnl, decimal_spec.asset_decimals
        ),
        unrealised_pnl=num_from_padded_int(
            position.unrealised_pnl, decimal_spec.asset_decimals
        ),
        average_entry_price=num_from_padded_int(
            position.average_entry_price, decimal_spec.price_decimals
        ),
        loss_socialisation_amount=num_from_padded_int(
            position.loss_socialisation_amount,
            decimal_spec.asset_decimals,
        ),
        updated_at=position.updated_at,
    )


def _transfer_from_proto(
    transfer: vega_protos.vega.Transfer, decimal_spec: DecimalSpec
) -> Transfer:
    return Transfer(
        id=transfer.id,
        party_from=getattr(transfer, "from"),
        from_account_type=transfer.from_account_type,
        party_to=transfer.to,
        to_account_type=transfer.to_account_type,
        asset=transfer.asset,
        amount=num_from_padded_int(transfer.amount, decimal_spec.asset_decimals),
        reference=transfer.reference,
        status=transfer.status,
        timestamp=transfer.timestamp,
        reason=transfer.reason,
        one_off=transfer.one_off,
        recurring=transfer.recurring,
    )


def _margin_estimate_from_proto(
    margin_estimate: data_node_protos_v2.trading_data.MarginEstimate,
    decimal_spec: DecimalSpec,
) -> MarginEstimate:
    return MarginEstimate(
        best_case=_margin_level_from_proto(
            margin_level=margin_estimate.best_case, decimal_spec=decimal_spec
        ),
        worst_case=_margin_level_from_proto(
            margin_level=margin_estimate.worst_case, decimal_spec=decimal_spec
        ),
    )


def _liquidation_estimate_from_proto(
    liquidation_estimate: data_node_protos_v2.trading_data.LiquidationEstimate,
):
    return LiquidationEstimate(
        best_case=_liquidation_price_from_proto(liquidation_estimate.best_case),
        worst_case=_liquidation_price_from_proto(liquidation_estimate.worst_case),
    )


def _liquidation_price_from_proto(
    liquidation_price: data_node_protos_v2.trading_data.LiquidationPrice,
):
    return LiquidationPrice(
        open_volume_only=float(liquidation_price.open_volume_only),
        including_buy_orders=float(liquidation_price.including_buy_orders),
        including_sell_orders=float(liquidation_price.including_sell_orders),
    )


def positions_by_market(
    data_client: vac.VegaTradingDataClientV2,
    pub_key: str,
    market_id: Optional[str] = None,
    market_price_decimals_map: Optional[Dict[str, int]] = None,
    market_position_decimals_map: Optional[Dict[str, int]] = None,
    market_to_asset_map: Optional[Dict[str, str]] = None,
    asset_decimals_map: Optional[Dict[str, int]] = None,
) -> Union[Dict[str, Position], Position]:
    """Output positions of a party."""

    raw_positions = data_raw.positions_by_market(
        pub_key=pub_key, market_id=market_id, data_client=data_client
    )
    if len(raw_positions) == 0:
        logging.debug(
            f"No positions to return for pub_key={pub_key}, market_id={market_id}"
        )
        return None

    positions = {}
    for pos in raw_positions:
        market_info = None

        # Update maps if value does not exist for current market id
        if pos.market_id not in market_price_decimals_map:
            if market_info is None:
                market_info = data_raw.market_info(
                    market_id=pos.market_id, data_client=data_client
                )
            market_price_decimals_map[pos.market_id] = int(market_info.decimal_places)
        if pos.market_id not in market_position_decimals_map:
            if market_info is None:
                market_info = data_raw.market_info(
                    market_id=pos.market_id, data_client=data_client
                )
            market_position_decimals_map[pos.market_id] = int(
                market_info.position_decimal_places
            )
        if pos.market_id not in market_to_asset_map:
            if market_info is None:
                market_info = data_raw.market_info(
                    market_id=pos.market_id, data_client=data_client
                )
            market_to_asset_map[
                pos.market_id
            ] = market_info.tradable_instrument.instrument.future.settlement_asset

        # Update maps if value does not exist for current asset id
        if market_to_asset_map[pos.market_id] not in asset_decimals_map:
            asset_info = data_raw.asset_info(
                asset_id=market_to_asset_map[pos.market_id],
                data_client=data_client,
            )
            asset_decimals_map[pos.market_id] = int(asset_info.details.decimals)

        # Convert raw proto into a market-sim Position object
        positions[pos.market_id] = _position_from_proto(
            position=pos,
            decimal_spec=DecimalSpec(
                price_decimals=market_price_decimals_map[pos.market_id],
                position_decimals=market_position_decimals_map[pos.market_id],
                asset_decimals=asset_decimals_map[market_to_asset_map[pos.market_id]],
            ),
        )

    if market_id is None:
        return positions
    else:
        return positions[market_id]


def _market_data_from_proto(
    market_data: vega_protos.vega.MarketData,
    decimal_spec: DecimalSpec,
):
    return MarketData(
        mark_price=num_from_padded_int(
            market_data.mark_price, decimal_spec.price_decimals
        ),
        best_bid_price=num_from_padded_int(
            market_data.best_bid_price, decimal_spec.price_decimals
        ),
        best_bid_volume=num_from_padded_int(
            market_data.best_bid_volume, decimal_spec.position_decimals
        ),
        best_offer_price=num_from_padded_int(
            market_data.best_offer_price, decimal_spec.price_decimals
        ),
        best_offer_volume=num_from_padded_int(
            market_data.best_offer_volume, decimal_spec.position_decimals
        ),
        best_static_bid_price=num_from_padded_int(
            market_data.best_static_bid_price, decimal_spec.price_decimals
        ),
        best_static_bid_volume=num_from_padded_int(
            market_data.best_static_bid_price, decimal_spec.position_decimals
        ),
        best_static_offer_price=num_from_padded_int(
            market_data.best_static_offer_price, decimal_spec.price_decimals
        ),
        best_static_offer_volume=num_from_padded_int(
            market_data.best_static_offer_volume, decimal_spec.position_decimals
        ),
        mid_price=num_from_padded_int(
            market_data.mid_price, decimal_spec.price_decimals
        ),
        static_mid_price=num_from_padded_int(
            market_data.static_mid_price, decimal_spec.price_decimals
        ),
        market_id=market_data.market,
        timestamp=market_data.timestamp,
        open_interest=market_data.open_interest,
        auction_end=market_data.auction_end,
        auction_start=market_data.auction_start,
        indicative_price=num_from_padded_int(
            market_data.indicative_price, decimal_spec.price_decimals
        ),
        indicative_volume=num_from_padded_int(
            market_data.indicative_volume, decimal_spec.price_decimals
        ),
        market_trading_mode=market_data.market_trading_mode,
        trigger=market_data.trigger,
        extension_trigger=market_data.extension_trigger,
        target_stake=num_from_padded_int(
            market_data.target_stake, decimal_spec.asset_decimals
        ),
        supplied_stake=num_from_padded_int(
            market_data.supplied_stake, decimal_spec.asset_decimals
        ),
        market_value_proxy=num_from_padded_int(
            market_data.market_value_proxy, decimal_spec.asset_decimals
        ),
        price_monitoring_bounds=_price_monitoring_bounds_from_proto(
            market_data.price_monitoring_bounds, decimal_spec.price_decimals
        ),
        liquidity_provider_fee_share=_liquidity_provider_fee_share_from_proto(
            market_data.liquidity_provider_fee_share,
            decimal_spec.asset_decimals,
        ),
        market_state=market_data.market_state,
        next_mark_to_market=market_data.next_mark_to_market,
        last_traded_price=num_from_padded_int(
            market_data.last_traded_price, decimal_spec.price_decimals
        ),
    )


def _price_monitoring_bounds_from_proto(
    price_monitoring_bounds,
    price_decimals: int,
) -> List[PriceMonitoringBounds]:
    return [
        PriceMonitoringBounds(
            min_valid_price=num_from_padded_int(
                individual_bound.min_valid_price,
                price_decimals,
            ),
            max_valid_price=num_from_padded_int(
                individual_bound.max_valid_price,
                price_decimals,
            ),
            trigger=individual_bound.trigger,
            reference_price=num_from_padded_int(
                individual_bound.reference_price,
                price_decimals,
            ),
        )
        for individual_bound in price_monitoring_bounds
    ]


def _liquidity_provider_fee_share_from_proto(
    liquidity_provider_fee_share,
    asset_decimals,
) -> List[PriceMonitoringBounds]:
    return [
        LiquidityProviderFeeShare(
            party=individual_liquidity_provider_fee_share.party,
            equity_like_share=float(
                individual_liquidity_provider_fee_share.equity_like_share
            ),
            average_entry_valuation=num_from_padded_int(
                float(individual_liquidity_provider_fee_share.average_entry_valuation),
                asset_decimals,
            ),
            average_score=float(
                individual_liquidity_provider_fee_share.equity_like_share
            ),
        )
        for individual_liquidity_provider_fee_share in liquidity_provider_fee_share
    ]


def list_accounts(
    data_client: vac.VegaTradingDataClientV2,
    pub_key: Optional[str] = None,
    asset_id: Optional[str] = None,
    market_id: Optional[str] = None,
    asset_decimals_map: Optional[Dict[str, int]] = None,
) -> List[AccountData]:
    """Output money in general accounts/margin accounts/bond accounts (if exists)
    of a party."""
    accounts = data_raw.list_accounts(
        data_client=data_client,
        party_id=pub_key,
        asset_id=asset_id,
        market_id=market_id,
    )

    asset_decimals_map = {} if asset_decimals_map is None else asset_decimals_map
    output_accounts = []
    for account in accounts:
        if account.asset not in asset_decimals_map:
            asset_decimals_map[account.asset] = get_asset_decimals(
                asset_id=account.asset,
                data_client=data_client,
            )

        output_accounts.append(
            AccountData(
                owner=account.owner,
                balance=num_from_padded_int(
                    int(account.balance), asset_decimals_map.setdefault(account.asset)
                ),
                asset=account.asset,
                type=account.type,
                market_id=account.market_id,
            )
        )
    return output_accounts


def party_account(
    pub_key: str,
    asset_id: str,
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
    asset_dp: Optional[int] = None,
) -> PartyMarketAccount:
    """Output money in general accounts/margin accounts/bond accounts (if exists)
    of a party."""
    accounts = data_raw.list_accounts(
        data_client=data_client,
        party_id=pub_key,
        asset_id=asset_id,
    )

    asset_dp = (
        asset_dp if asset_dp is not None else get_asset_decimals(asset_id, data_client)
    )

    general, margin, bond = 0, 0, 0  # np.nan, np.nan, np.nan
    for account in accounts:
        if (
            account.market_id
            and account.market_id != "!"
            and account.market_id != market_id
        ):
            # The 'general' account type has no market ID, so we have to pull
            # all markets then filter down here
            continue
        if account.type == vega_protos.vega.ACCOUNT_TYPE_GENERAL:
            general = num_from_padded_int(float(account.balance), asset_dp)
        if account.type == vega_protos.vega.ACCOUNT_TYPE_MARGIN:
            margin = num_from_padded_int(float(account.balance), asset_dp)
        if account.type == vega_protos.vega.ACCOUNT_TYPE_BOND:
            bond = num_from_padded_int(float(account.balance), asset_dp)

    return PartyMarketAccount(general, margin, bond)


def find_market_id(
    name: str, data_client: vac.VegaTradingDataClientV2, raise_on_missing: bool = False
) -> str:
    """Looks up the Market ID of a given market name

    Args:
        symbol:
            str, The symbol of the asset to look up
        data_client:
            VegaTradingDataClientV2, the gRPC data client
        raise_on_missing:
            bool, whether to raise an Error or silently return
                if the asset does not exist

    Returns:
        str, the ID of the asset
    """
    markets = data_raw.all_markets(data_client=data_client)

    acceptable_states = [
        vega_protos.markets.Market.STATE_PENDING,
        vega_protos.markets.Market.STATE_ACTIVE,
        vega_protos.markets.Market.STATE_SUSPENDED,
    ]

    market_ids = {}

    for market in markets:
        if market.tradable_instrument.instrument.name == name:
            if market.state in acceptable_states:
                market_ids[market.id] = market.market_timestamps.pending

    if len(market_ids) > 0:
        return max(market_ids, key=market_ids.get)

    if raise_on_missing:
        raise MissingMarketError(
            f"{name} market not found on specified Vega network, "
            + "please propose and create this market first"
        )


def find_asset_id(
    symbol: str,
    data_client: vac.VegaTradingDataClientV2,
    enabled: bool = True,
    raise_on_missing: bool = False,
) -> str:
    """Looks up the Asset ID of a given asset name

    Args:
        symbol:
            str, The symbol of the asset to look up
        data_client:
            VegaTradingDataClientV2, the gRPC data client
        raise_on_missing:
            bool, whether to raise an Error or silently return
                if the asset does not exist

    Returns:
        str, the ID of the asset
    """
    assets = data_raw.list_assets(data_client=data_client)
    # Find settlement asset
    for asset in assets:
        if asset.details.symbol == symbol:
            if (enabled) and (asset.status != vega_protos.assets.Asset.STATUS_ENABLED):
                continue
            return asset.id
    if raise_on_missing:
        raise MissingAssetError(
            f"{symbol} asset not found on specified Vega network, "
            + "please propose and create this asset first"
        )


def market_price_decimals(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> int:
    """Returns the number of decimal places a specified market uses for price units.

    Args:
        market_id:
            str, The ID of the market requested
        data_client:
            VegaTradingDataClientV2, an instantiated gRPC data client

    Returns:
        int, The number of decimal places the market uses
    """
    res = data_raw.market_info(
        market_id=market_id, data_client=data_client
    ).decimal_places
    return res


def market_position_decimals(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> int:
    """Returns the number of decimal places a specified market uses for position units.

    Args:
        market_id:
            str, The ID of the market requested
        data_client:
            VegaTradingDataClientV2,
    Returns:
        int, The number of decimal places the market uses for positions
    """
    return data_raw.market_info(
        market_id=market_id, data_client=data_client
    ).position_decimal_places


def get_asset_decimals(
    asset_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> int:
    """Returns the number of decimal places a specified asset uses for price.

    Args:
        asset_id:
            str, The ID of the asset requested
        data_client:
            VegaTradingDataClientV2, an instantiated gRPC data client

    Returns:
        int, The number of decimal places the asset uses
    """
    return data_raw.asset_info(
        asset_id=asset_id, data_client=data_client
    ).details.decimals


def open_orders_by_market(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
    price_decimals: Optional[int] = None,
    position_decimals: Optional[int] = None,
) -> OrdersBySide:
    """
    Output all active limit orders in current market.

    Args:
        market_id:
            str, ID for the market to load
        data_client:
            VegaTradingDataClientV2, instantiated gRPC client

    Returns:
        OrdersBySide, Live orders segregated by side
    """
    bids = []
    asks = []
    orders = all_orders(
        market_id=market_id,
        data_client=data_client,
        price_decimals=price_decimals,
        position_decimals=position_decimals,
        open_only=True,
    )
    for order in orders:
        (
            bids.append(order)
            if order.side == vega_protos.vega.SIDE_BUY
            else asks.append(order)
        )

    return OrdersBySide(bids, asks)


def list_orders(
    data_client: vac.VegaTradingDataClientV2,
    market_id: Optional[str] = None,
    party_id: Optional[str] = None,
    reference: Optional[str] = None,
    price_decimals: Optional[int] = None,
    position_decimals: Optional[int] = None,
    live_only: Optional[bool] = True,
) -> List[Order]:
    """Return a list of converted orders for the specified market, party, and reference.

    Args:
        data_client (vac.VegaTradingDataClientV2):
            An instantiated gRPC trading_data_client_v2.
        market_id (str):
            Id of market to return orders for.
        party_id (str):
            Id of party to return orders for.
        reference (str):
            References to return orders for.
        price_decimals (Optional[int], optional):
            Market price decimal places. Defaults to None.
        position_decimals (Optional[int], optional):
            Market position decimal places. Defaults to None.
        live_only (Optional[bool], optional):
            Whether to only return live orders. Defaults to True.

    Returns:
        List[Order]:
            A list of converted orders.
    """
    orders = data_raw.list_orders(
        data_client=data_client,
        party_id=party_id,
        market_id=market_id,
        reference=reference,
        live_only=live_only,
    )

    mkt_price_dp = (
        DefaultDict(lambda: price_decimals) if price_decimals is not None else {}
    )
    mkt_pos_dp = (
        DefaultDict(lambda: position_decimals) if position_decimals is not None else {}
    )

    output_orders = []
    for order in orders:
        if price_decimals is None and order.market_id not in mkt_price_dp:
            mkt_pos_dp[order.market_id] = market_position_decimals(
                market_id=order.market_id, data_client=data_client
            )
            mkt_price_dp[order.market_id] = market_price_decimals(
                market_id=order.market_id, data_client=data_client
            )

        converted_order = _order_from_proto(
            order,
            decimal_spec=DecimalSpec(
                price_decimals=mkt_price_dp[order.market_id],
                position_decimals=mkt_pos_dp[order.market_id],
            ),
        )
        output_orders.append(converted_order)
    return output_orders


def all_orders(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
    price_decimals: Optional[int] = None,
    position_decimals: Optional[int] = None,
    open_only: bool = False,
) -> OrdersBySide:
    """
    Output all active limit orders in current market.

    Args:
        market_id:
            str, ID for the market to load
        data_client:
            VegaTradingDataClientV2, instantiated gRPC client
        open_only:
            bool, default False, whether to only return still
                open orders

    Returns:
        OrdersBySide, Live orders segregated by side
    """
    orders = data_raw.list_orders(
        market_id=market_id,
        data_client=data_client,
    )

    mkt_price_dp = (
        DefaultDict(lambda: price_decimals) if price_decimals is not None else {}
    )
    mkt_pos_dp = (
        DefaultDict(lambda: position_decimals) if position_decimals is not None else {}
    )

    output_orders = []
    for order in orders:
        if price_decimals is None and order.market_id not in mkt_price_dp:
            mkt_pos_dp[order.market_id] = market_position_decimals(
                market_id=order.market_id, data_client=data_client
            )
            mkt_price_dp[order.market_id] = market_price_decimals(
                market_id=order.market_id, data_client=data_client
            )

        if open_only and order.status != vega_protos.vega.Order.Status.STATUS_ACTIVE:
            continue
        converted_order = _order_from_proto(
            order,
            DecimalSpec(
                price_decimals=mkt_price_dp[order.market_id],
                position_decimals=mkt_pos_dp[order.market_id],
            ),
        )
        output_orders.append(converted_order)
    return output_orders


def order_book_by_market(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> OrderBook:
    """
    Output state of order book for a given market.
    """

    orders = data_raw.list_orders(
        market_id=market_id,
        data_client=data_client,
    )

    mkt_price_dp = market_price_decimals(market_id=market_id, data_client=data_client)
    mkt_pos_dp = market_position_decimals(market_id=market_id, data_client=data_client)

    bids = {}
    asks = {}

    for order in orders:
        if order.status == vega_protos.vega.Order.Status.STATUS_ACTIVE:
            if order.side == vega_protos.vega.Side.SIDE_BUY:
                price = num_from_padded_int(order.price, mkt_price_dp)
                volume = num_from_padded_int(order.remaining, mkt_pos_dp)
                bids[price] = bids.get(price, 0) + volume
            else:
                price = num_from_padded_int(order.price, mkt_price_dp)
                volume = num_from_padded_int(order.remaining, mkt_pos_dp)
                asks[price] = bids.get(price, 0) + volume

    return OrderBook(bids, asks)


def market_account(
    market_id: str,
    account_type: vega_protos.vega.AccountType,
    data_client: vac.VegaTradingDataClientV2,
) -> float:
    """
    Returns the current asset value in the Market's fee account

    Args:
        market_id:
            str, The ID of the market to check
        account_type:
            vega.AccountType, the account type to check for

    Returns:
        float, the current balance in the market's fee asset
    """
    accounts = data_raw.list_accounts(
        asset_id=None, market_id=market_id, data_client=data_client
    )
    acct = {account.type: account for account in accounts}[account_type]

    asset_dp = get_asset_decimals(asset_id=acct.asset, data_client=data_client)
    return num_from_padded_int(
        acct.balance,
        asset_dp,
    )


def market_depth(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
    max_depth: Optional[int] = None,
    price_decimals: Optional[int] = None,
    position_decimals: Optional[int] = None,
) -> Optional[MarketDepth]:
    mkt_depth = data_raw.market_depth(
        data_client=data_client,
        market_id=market_id,
        max_depth=max_depth,
    )
    if mkt_depth is None:
        return mkt_depth

    mkt_price_dp = (
        price_decimals
        if price_decimals is not None
        else market_price_decimals(market_id=market_id, data_client=data_client)
    )
    mkt_pos_dp = (
        position_decimals
        if position_decimals is not None
        else market_position_decimals(market_id=market_id, data_client=data_client)
    )

    def _price_level_from_raw(level) -> PriceLevel:
        return PriceLevel(
            price=num_from_padded_int(
                level.price,
                mkt_price_dp,
            ),
            number_of_orders=level.number_of_orders,
            volume=num_from_padded_int(
                level.volume,
                mkt_pos_dp,
            ),
        )

    return MarketDepth(
        buys=[_price_level_from_raw(level) for level in mkt_depth.buy],
        sells=[_price_level_from_raw(level) for level in mkt_depth.sell],
    )


def has_liquidity_provision(
    data_client: vac.VegaTradingDataClientV2,
    market_id: str,
    party_id: str,
) -> bool:
    lip = data_raw.liquidity_provisions(
        data_client=data_client, market_id=market_id, party_id=party_id
    )
    return (
        lip
        and len(lip) > 0
        and lip[0].status
        in [
            vega_protos.vega.LiquidityProvision.Status.STATUS_ACTIVE,
            vega_protos.vega.LiquidityProvision.Status.STATUS_UNDEPLOYED,
            vega_protos.vega.LiquidityProvision.Status.STATUS_PENDING,
        ]
    )


def margin_levels(
    data_client: vac.VegaTradingDataClientV2,
    party_id: str,
    market_id: Optional[str] = None,
) -> List[MarginLevels]:
    asset_dp = {}
    margins = data_raw.margin_levels(
        data_client=data_client, party_id=party_id, market_id=market_id
    )
    res_margins = []
    for margin in margins:
        if margin.asset not in asset_dp:
            asset_dp[margin.asset] = get_asset_decimals(
                asset_id=margin.asset, data_client=data_client
            )
        res_margins.append(
            _margin_level_from_proto(
                margin, DecimalSpec(asset_decimals=asset_dp[margin.asset])
            )
        )
    return res_margins


def get_trades(
    data_client: vac.VegaTradingDataClientV2,
    market_id: str,
    party_id: Optional[str] = None,
    order_id: Optional[str] = None,
    market_price_decimals_map: Optional[Dict[str, int]] = None,
    market_position_decimals_map: Optional[Dict[str, int]] = None,
    market_asset_decimals_map: Optional[Dict[str, int]] = None,
    start: Optional[datetime.datetime] = None,
    end: Optional[datetime.datetime] = None,
) -> List[Trade]:
    base_trades = data_raw.get_trades(
        data_client=data_client,
        party_id=party_id,
        market_id=market_id,
        order_id=order_id,
        start=start,
        end=end,
    )
    market_price_decimals_map = market_price_decimals_map or {}
    market_position_decimals_map = market_position_decimals_map or {}
    market_asset_decimals_map = market_asset_decimals_map or {}

    res_trades = []
    for trade in base_trades:
        if trade.market_id not in market_price_decimals_map:
            market_price_decimals_map[trade.market_id] = market_price_decimals(
                market_id=trade.market_id, data_client=data_client
            )
        if trade.market_id not in market_position_decimals_map:
            market_position_decimals_map[trade.market_id] = market_position_decimals(
                market_id=trade.market_id, data_client=data_client
            )
        if trade.market_id not in market_asset_decimals_map:
            market_asset_decimals_map[trade.market_id] = get_asset_decimals(
                asset_id=data_raw.market_info(
                    market_id=market_id, data_client=data_client
                ).tradable_instrument.instrument.future.settlement_asset,
                data_client=data_client,
            )
        res_trades.append(
            _trade_from_proto(
                trade,
                DecimalSpec(
                    price_decimals=market_price_decimals_map[trade.market_id],
                    position_decimals=market_position_decimals_map[trade.market_id],
                    asset_decimals=market_asset_decimals_map[trade.market_id],
                ),
            )
        )
    return res_trades


def ping(data_client: vac.VegaTradingDataClientV2):
    return data_client.Ping(data_node_protos_v2.trading_data.PingRequest())


def list_transfers(
    data_client: vac.VegaTradingDataClientV2,
    party_id: Optional[str] = None,
    direction: data_node_protos_v2.trading_data.TransferDirection = None,
) -> List[Transfer]:
    """Returns a list of processed transfers.

    Args:
        data_client (vac.VegaTradingDataClientV2):
            An instantiated gRPC trading data client
        party_id (Optional[str], optional):
            Public key for the specified party. Defaults to None.
        direction (Optional[data_node_protos_v2.trading_data.TransferDirection], optional):
            Direction of transfers to return. Defaults to None.

    Returns:
        List[Transfer]:
            A list of processed Transfer objects for the specified party and direction.
    """

    transfers = data_raw.list_transfers(
        data_client=data_client,
        party_id=party_id,
        direction=direction,
    )

    asset_dp = {}
    res_transfers = []

    for transfer in transfers:
        if transfer.status == events_protos.Transfer.Status.STATUS_REJECTED:
            continue

        if transfer.asset not in asset_dp:
            asset_dp[transfer.asset] = get_asset_decimals(
                asset_id=transfer.asset, data_client=data_client
            )
        res_transfers.append(
            _transfer_from_proto(
                transfer=transfer,
                decimal_spec=DecimalSpec(asset_decimals=asset_dp[transfer.asset]),
            )
        )

    return res_transfers


def _stream_handler(
    stream_item: vega_protos.api.v1.core.ObserveEventBusResponse,
    extraction_fn: Callable[[vega_protos.events.v1.events.BusEvent], T],
    conversion_fn: Callable[[T, int, int, int], S],
    mkt_pos_dp: Optional[Dict[str, int]] = None,
    mkt_price_dp: Optional[Dict[str, int]] = None,
    mkt_to_asset: Optional[Dict[str, str]] = None,
    asset_dp: Optional[Dict[str, int]] = None,
) -> S:
    mkt_pos_dp = mkt_pos_dp if mkt_pos_dp is not None else {}
    mkt_price_dp = mkt_price_dp if mkt_price_dp is not None else {}
    mkt_to_asset = mkt_to_asset if mkt_to_asset is not None else {}
    asset_dp = asset_dp if asset_dp is not None else {}

    event = extraction_fn(stream_item)

    market_id = getattr(event, "market_id", getattr(event, "market", None))

    # Check market creation event observed
    if (market_id is not None) and (
        (market_id not in mkt_pos_dp)
        or (market_id not in mkt_price_dp)
        or (market_id not in mkt_to_asset)
    ):
        return

    asset_id = getattr(
        event,
        "asset",
        mkt_to_asset[market_id] if market_id is not None else None,
    )

    # Check asset creation event observed
    if (asset_id is not None) and (asset_id not in asset_dp):
        return

    return conversion_fn(
        event,
        DecimalSpec(
            price_decimals=mkt_price_dp[market_id] if market_id is not None else None,
            position_decimals=mkt_pos_dp[market_id] if market_id is not None else None,
            asset_decimals=asset_dp[asset_id],
        ),
    )


def list_ledger_entries(
    data_client: vac.VegaTradingDataClientV2,
    close_on_account_filters: bool = False,
    asset_id: Optional[str] = None,
    from_party_ids: Optional[List[str]] = None,
    to_party_ids: Optional[List[str]] = None,
    from_account_types: Optional[list[vega_protos.vega.AccountType]] = None,
    from_market_ids: Optional[List[str]] = None,
    to_market_ids: Optional[List[str]] = None,
    to_account_types: Optional[list[vega_protos.vega.AccountType]] = None,
    transfer_types: Optional[list[vega_protos.vega.TransferType]] = None,
    from_datetime: Optional[datetime.datetime] = None,
    to_datetime: Optional[datetime.datetime] = None,
    asset_decimals_map: Optional[Dict[str, int]] = None,
) -> List[AggregatedLedgerEntry]:
    """Returns a list of ledger entries matching specific filters as provided.
    These detail every transfer of funds between accounts within the Vega system,
    including fee/rewards payments and transfers between user margin/general/bond
    accounts.

    Note: At least one of the from_*/to_* filters, or asset ID, must be specified.

    Args:
        data_client:
            vac.VegaTradingDataClientV2, An instantiated gRPC trading data client
        close_on_account_filters:
            bool, default False, Whether both 'from' and 'to' filters must both match
                a given transfer for inclusion. If False, entries matching either
                'from' or 'to' will also be included.
        asset_id:
            Optional[str], filter to only transfers of specific asset ID
        from_party_ids:
            Optional[List[str]], Only include transfers from specified parties
        from_market_ids:
            Optional[List[str]], Only include transfers from specified markets
        from_account_types:
            Optional[List[str]], Only include transfers from specified account types
        to_party_ids:
            Optional[List[str]], Only include transfers to specified parties
        to_market_ids:
            Optional[List[str]], Only include transfers to specified markets
        to_account_types:
            Optional[List[str]], Only include transfers to specified account types
        transfer_types:
            Optional[List[vega_protos.vega.AccountType]], Only include transfers
                of specified types
        from_datetime:
            Optional[datetime.datetime], Only include transfers occurring after
                this time
        to_datetime:
            Optional[datetime.datetime], Only include transfers occurring before
                this time
    Returns:
        List[AggregatedLedgerEntry]
            A list of all transfers matching the requested criteria
    """
    raw_ledger_entries = data_raw.list_ledger_entries(
        data_client=data_client,
        close_on_account_filters=close_on_account_filters,
        asset_id=asset_id,
        from_party_ids=from_party_ids,
        to_party_ids=to_party_ids,
        from_account_types=from_account_types,
        from_market_ids=from_market_ids,
        to_market_ids=to_market_ids,
        transfer_types=transfer_types,
        from_datetime=from_datetime,
        to_datetime=to_datetime,
        to_account_types=to_account_types,
    )

    asset_decimals_map = {} if asset_decimals_map is None else asset_decimals_map
    ledger_entries = []
    for entry in raw_ledger_entries:
        if entry.asset_id not in asset_decimals_map:
            asset_decimals_map[entry.asset_id] = get_asset_decimals(
                asset_id=entry.asset_id,
                data_client=data_client,
            )
        ledger_entries.append(
            _aggregated_ledger_entry_from_proto(
                entry, DecimalSpec(asset_decimals=asset_decimals_map[entry.asset_id])
            )
        )
    return ledger_entries


def trades_subscription_handler(
    stream: Iterable[vega_protos.api.v1.core.ObserveEventBusResponse],
    mkt_pos_dp: Optional[Dict[str, int]] = None,
    mkt_price_dp: Optional[Dict[str, int]] = None,
    mkt_to_asset: Optional[Dict[str, str]] = None,
    asset_dp: Optional[Dict[str, int]] = None,
) -> Trade:
    """Subscribe to a stream of Order updates from the data-node.
    The stream of orders returned from this function is an iterable which
    does not end and will continue to tick another order update whenever
    one is received.

    Returns:
        Iterable[Trade], Infinite iterable of trade updates
    """
    return _stream_handler(
        stream_item=stream,
        extraction_fn=lambda evt: evt.trade,
        conversion_fn=_trade_from_proto,
        mkt_pos_dp=mkt_pos_dp,
        mkt_price_dp=mkt_price_dp,
        mkt_to_asset=mkt_to_asset,
        asset_dp=asset_dp,
    )


def order_subscription_handler(
    stream: Iterable[vega_protos.api.v1.core.ObserveEventBusResponse],
    mkt_pos_dp: Optional[Dict[str, int]] = None,
    mkt_price_dp: Optional[Dict[str, int]] = None,
    mkt_to_asset: Optional[Dict[str, str]] = None,
    asset_dp: Optional[Dict[str, int]] = None,
) -> Order:
    """Subscribe to a stream of Order updates from the data-node.
    The stream of orders returned from this function is an iterable which
    does not end and will continue to tick another order update whenever
    one is received.

    Args:
        market_id:
            Optional[str], If provided, only update orders from this market
        party_id:
            Optional[str], If provided, only update orders from this party
    Returns:
        Iterable[Order], Infinite iterable of order updates
    """
    return _stream_handler(
        stream_item=stream,
        extraction_fn=lambda evt: evt.order,
        conversion_fn=_order_from_proto,
        mkt_pos_dp=mkt_pos_dp,
        mkt_price_dp=mkt_price_dp,
        mkt_to_asset=mkt_to_asset,
        asset_dp=asset_dp,
    )


def transfer_subscription_handler(
    stream: Iterable[vega_protos.api.v1.core.ObserveEventBusResponse],
    mkt_pos_dp: Optional[Dict[str, int]] = None,
    mkt_price_dp: Optional[Dict[str, int]] = None,
    mkt_to_asset: Optional[Dict[str, str]] = None,
    asset_dp: Optional[Dict[str, int]] = None,
) -> Transfer:
    return _stream_handler(
        stream_item=stream,
        extraction_fn=lambda evt: evt.transfer,
        conversion_fn=_transfer_from_proto,
        mkt_pos_dp=mkt_pos_dp,
        mkt_price_dp=mkt_price_dp,
        mkt_to_asset=mkt_to_asset,
        asset_dp=asset_dp,
    )


def ledger_entries_subscription_handler(
    stream_item: vega_protos.api.v1.core.ObserveEventBusResponse,
    asset_dp: Optional[Dict[str, int]] = None,
) -> Iterable[LedgerEntry]:
    ledger_entries = []
    for ledger_movement in stream_item.ledger_movements.ledger_movements:
        for ledger_entry in ledger_movement.entries:
            asset_id = ledger_entry.from_account.asset_id
            ledger_entries.append(
                _ledger_entry_from_proto(
                    ledger_entry,
                    asset_decimals=asset_dp[asset_id],
                )
            )
    return ledger_entries


def market_data_subscription_handler(
    stream_item: vega_protos.api.v1.core.ObserveEventBusResponse,
    mkt_pos_dp: Optional[Dict[str, int]] = None,
    mkt_price_dp: Optional[Dict[str, int]] = None,
    mkt_to_asset: Optional[Dict[str, str]] = None,
    asset_dp: Optional[Dict[str, int]] = None,
):
    return _stream_handler(
        stream_item=stream_item,
        extraction_fn=lambda evt: evt.market_data,
        conversion_fn=_market_data_from_proto,
        mkt_pos_dp=mkt_pos_dp,
        mkt_price_dp=mkt_price_dp,
        mkt_to_asset=mkt_to_asset,
        asset_dp=asset_dp,
    )


def get_latest_market_data(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
    market_price_decimals_map: Optional[Dict[str, int]] = None,
    market_position_decimals_map: Optional[Dict[str, int]] = None,
    market_to_asset_map: Optional[Dict[str, str]] = None,
    asset_decimals_map: Optional[Dict[str, int]] = None,
) -> MarketData:
    # Get latest market data
    market_data = data_raw.get_latest_market_data(
        market_id=market_id, data_client=data_client
    )

    market_price_decimals_map = (
        market_price_decimals_map if market_price_decimals_map is not None else {}
    )
    market_position_decimals_map = (
        market_position_decimals_map if market_position_decimals_map is not None else {}
    )
    market_to_asset_map = market_to_asset_map if market_to_asset_map is not None else {}
    asset_decimals_map = asset_decimals_map if asset_decimals_map is not None else {}

    if market_id not in market_price_decimals_map:
        market_price_decimals_map[market_id] = market_price_decimals(
            market_id=market_id, data_client=data_client
        )
    if market_id not in market_position_decimals_map:
        market_position_decimals_map[market_id] = market_position_decimals(
            market_id=market_id, data_client=data_client
        )
    if market_id not in market_to_asset_map:
        market_to_asset_map[market_id] = data_raw.market_info(
            market_id=market_id, data_client=data_client
        ).tradable_instrument.instrument.future.settlement_asset
    if market_to_asset_map[market_id] not in asset_decimals_map:
        asset_decimals_map[market_to_asset_map[market_id]] = get_asset_decimals(
            asset_id=market_to_asset_map[market_id],
            data_client=data_client,
        )
    # Convert from proto
    return _market_data_from_proto(
        market_data=market_data,
        decimal_spec=DecimalSpec(
            price_decimals=market_price_decimals_map[market_data.market],
            position_decimals=market_position_decimals_map[market_data.market],
            asset_decimals=asset_decimals_map[market_to_asset_map[market_data.market]],
        ),
    )


def get_risk_factors(
    data_client: vac.VegaTradingDataClientV2,
    market_id: str,
):
    raw_risk_factors = data_raw.get_risk_factors(
        data_client=data_client, market_id=market_id
    )
    return RiskFactor(
        market_id=market_id,
        short=float(raw_risk_factors.risk_factor.short),
        long=float(raw_risk_factors.risk_factor.long),
    )


def estimate_position(
    data_client: vac.VegaTradingDataClientV2,
    market_id: str,
    open_volume: int,
    orders: Optional[List[Tuple[str, str, int, bool]]] = None,
    collateral_available: Optional[str] = None,
    asset_decimals: Optional[Dict[str, int]] = {},
) -> Tuple[MarginEstimate, LiquidationEstimate,]:
    if orders is not None:
        proto_orders = [
            data_node_protos_v2.trading_data.OrderInfo(
                side=order[0],
                price=order[1],
                remaining=order[2],
                is_market_order=order[3],
            )
            for order in orders
        ]

    margin_estimate, liquidation_estimate = data_raw.estimate_position(
        data_client=data_client,
        market_id=market_id,
        open_volume=open_volume,
        orders=proto_orders if orders is not None else None,
        collateral_available=collateral_available,
    )

    if margin_estimate.best_case.asset not in asset_decimals:
        asset_decimals[margin_estimate.best_case.asset] = get_asset_decimals(
            asset_id=margin_estimate.best_case.market_id, data_client=data_client
        )

    converted_margin_estimate = _margin_estimate_from_proto(
        margin_estimate=margin_estimate,
        decimal_spec=DecimalSpec(
            asset_decimals=asset_decimals[margin_estimate.best_case.asset]
        ),
    )

    converted_liquidation_estimate = _liquidation_estimate_from_proto(
        liquidation_estimate=liquidation_estimate,
    )

    return converted_margin_estimate, converted_liquidation_estimate


def get_stake(
    data_client: vac.trading_data_grpc_v2, party_id: str, asset_decimals: int
):
    return num_from_padded_int(
        data_raw.get_stake(data_client=data_client, party_id=party_id),
        decimals=asset_decimals,
    )


def get_asset(
    data_client: vac.trading_data_grpc_v2,
    asset_id: str,
):
    return data_raw.asset_info(data_client=data_client, asset_id=asset_id)



import logging
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Type

from gymnasium import spaces
from pettingzoo.utils.env import AgentID, ParallelEnv
from pettingzoo.test import api_test

from vega_sim.reinforcement.v2.agents.puppets import (
    AGENT_TYPE_TO_ACTION,
    AgentType,
    ForcedSide,
    Side,
)
from vega_sim.reinforcement.v2.learning_environment import Environment
from vega_sim.reinforcement.v2.rewards import REWARD_ENUM_TO_CLASS, Reward
from vega_sim.reinforcement.v2.stable_baselines.states import (
    position_state_with_fees_obs_space,
    price_state_with_fees_obs_space,
)
from vega_sim.reinforcement.v2.states import PositionOnly, PriceStateWithFees, State
from vega_sim.scenario.registry import CurveMarketMaker

logger = logging.getLogger(__name__)


class ActionType(Enum):
    MARKET = "market"
    AT_TOUCH_ONE_SIDE = "at_touch_one_side"


ACTION_TO_AGENT = {
    ActionType.MARKET: AgentType.MARKET_ORDER,
    ActionType.AT_TOUCH_ONE_SIDE: AgentType.AT_TOUCH,
}


@dataclass
class AgentConfig:
    action_type: ActionType
    state_type: Type[State]
    reward_type: Optional[Reward] = None
    terminal_reward_type: Optional[Reward] = None


class MultiAgentVegaEnv(ParallelEnv):
    """Custom Environment that follows gym interface."""

    metadata = {"render.modes": ["human"]}

    def __init__(
        self,
        agents: List[AgentConfig],
        num_levels_state: int = 5,
        trade_volume: float = 1,
        steps_per_trading_session: int = 1000,
        unified_reward: Optional[Reward] = None,
    ):
        super().__init__()
        self.num_levels_state = num_levels_state
        self.agent_configs = agents
        self.trade_volume = trade_volume
        self.steps_per_trading_session = steps_per_trading_session
        self.current_step = 0
        self.learner_names = {
            str(i): f"learner_{i}" for i in range(len(self.agent_configs))
        }
        self.unified_reward = unified_reward

        # Define action and observation space
        # They must be gym.spaces objects
        # Example when using discrete actions:
        self.action_spaces = {
            str(i): self._get_action_space(action_type=agent_config.action_type)
            for (i, agent_config) in enumerate(self.agent_configs)
        }
        # Example for using image as input (channel-first; channel-last also works):
        self.observation_spaces = {
            str(i): self._get_observation_space(state_type=agent_config.state_type)
            for (i, agent_config) in enumerate(self.agent_configs)
        }

        scenario = CurveMarketMaker(
            market_decimal=3,
            asset_decimal=5,
            market_position_decimal=2,
            initial_price=1000.0,
            lp_commitamount=100000,
            initial_asset_mint=1e8,
            step_length_seconds=1,
            block_length_seconds=1,
            buy_intensity=5,
            sell_intensity=5,
            market_name="ETH",
            num_steps=self.steps_per_trading_session * 2,
            random_agent_ordering=False,
            sigma=100,
            asset_name="DAI",
        )

        self.env = Environment(
            agents={
                learner_name: ACTION_TO_AGENT[self.agent_configs[int(i)].action_type]
                for (i, learner_name) in self.learner_names.items()
            },
            agent_to_reward=(
                {
                    learner_name: self.agent_configs[int(i)].reward_type
                    for (i, learner_name) in self.learner_names.items()
                }
                if self.unified_reward is None
                else self.unified_reward
            ),
            agent_to_state={
                learner_name: self.agent_configs[int(i)].state_type
                for (i, learner_name) in self.learner_names.items()
            },
            scenario=scenario,
        )
        self.agents = list(self.learner_names.keys())
        self.possible_agents = self.agents

    def _get_action_space(self, action_type: ActionType) -> spaces.Space:
        if action_type == ActionType.MARKET:
            return spaces.Discrete(3)
        elif action_type == ActionType.AT_TOUCH_ONE_SIDE:
            return spaces.Discrete(2)
        else:
            raise Exception(f"Action type {action_type} is not implemented")

    def _get_observation_space(self, state_type: Type[State]) -> spaces.Space:
        if state_type == PriceStateWithFees:
            return price_state_with_fees_obs_space(num_levels=self.num_levels_state)
        if state_type == PositionOnly:
            return position_state_with_fees_obs_space()

    def _action_conversion(self, agent, action):
        agent_config = self.agent_configs[int(agent)]
        if agent_config.action_type == ActionType.MARKET:
            return AGENT_TYPE_TO_ACTION[AgentType.MARKET_ORDER](
                side=Side(action), volume=self.trade_volume
            )
        elif agent_config.action_type == ActionType.AT_TOUCH_ONE_SIDE:
            return AGENT_TYPE_TO_ACTION[AgentType.AT_TOUCH](
                side=ForcedSide(action), volume=self.trade_volume
            )
        else:
            raise Exception(
                f"Action type {agent_config.action_type} is not implemented"
            )

    def step(self, actions):
        step_res = self.env.step(
            {
                self.learner_names[agent]: self._action_conversion(
                    action=action, agent=agent
                )
                for (agent, action) in actions.items()
            }
        )
        self.current_step += 1

        is_terminal = self.current_step >= self.steps_per_trading_session

        self.latest_observations = {}
        self.latest_rewards = {}
        self.latest_terminations = {}
        self.latest_truncations = {}
        self.latest_infos = {}

        for agent_id, agent_name in self.learner_names.items():
            agent_step_res = step_res[agent_name]
            self.latest_observations[agent_id] = agent_step_res.observation.to_array()

            reward = agent_step_res.reward
            if (
                is_terminal
                and self.agent_configs[int(agent_id)].terminal_reward_type is not None
            ):
                terminal_reward = REWARD_ENUM_TO_CLASS[
                    self.agent_configs[int(agent_id)].terminal_reward_type
                ](
                    agent_key=agent_name,
                    asset_id=self.env._asset_id,
                    market_id=self.env._market_id,
                )
                reward += self.env.calculate_reward(terminal_reward)
            self.latest_rewards[str(agent_id)] = reward
            self.latest_terminations[str(agent_id)] = is_terminal
            self.latest_truncations[str(agent_id)] = False
            self.latest_infos[str(agent_id)] = {}

        return (
            self.latest_observations,
            self.latest_rewards,
            self.latest_terminations,
            self.latest_truncations,
            self.latest_infos,
        )

    def reset(self, seed=None, options=None):
        self.env.reset()
        step_res = self.env.step(
            {learner_name: None for learner_name in self.learner_names.values()}
        )

        self.current_step = 0
        return {
            i: step_res[name].observation.to_array()
            for (i, name) in self.learner_names.items()
        }, {i: {} for i in self.learner_names.keys()}

    def render(self, mode="human"):
        pass

    def close(self):
        self.env.stop()

    def action_space(self, agent: AgentID):
        return self.action_spaces[agent]

    def observation_space(self, agent: AgentID):
        return self.observation_spaces[agent]


if __name__ == "__main__":
    from pettingzoo.utils.conversions import parallel_to_aec_wrapper

    env = MultiAgentVegaEnv(
        agents=[
            AgentConfig(
                action_type=ActionType.AT_TOUCH_ONE_SIDE,
                state_type=PriceStateWithFees,
                reward_type=Reward.PNL,
                terminal_reward_type=Reward.SQ_INVENTORY_PENALTY,
            )
        ],
        steps_per_trading_session=200,
    )
    api_test(parallel_to_aec_wrapper(env), num_cycles=1, verbose_progress=True)



from abc import abstractmethod
from dataclasses import dataclass
from typing import Optional
import numpy as np

from vega_sim.proto.vega import markets as markets_protos
from vega_sim.service import VegaService


@dataclass(frozen=True)
class State:
    @classmethod
    @abstractmethod
    def from_vega(
        cls, vega: VegaService, for_key_name: str, for_wallet_name: Optional[str] = None
    ):
        return cls()

    @abstractmethod
    def to_array(self) -> np.array:
        pass


@dataclass(frozen=True)
class SimpleState(State):
    balance: float
    position: float
    best_bid: float
    best_ask: float

    @classmethod
    def from_vega(
        cls,
        vega: VegaService,
        for_key_name: str,
        market_id: str,
        asset_id: str,
        for_wallet_name: Optional[str] = None,
    ) -> "SimpleState":
        balance = sum(
            vega.party_account(
                for_key_name,
                asset_id=asset_id,
                market_id=market_id,
                wallet_name=for_wallet_name,
            )
        )
        position = vega.positions_by_market(
            key_name=for_key_name, wallet_name=for_wallet_name, market_id=market_id
        )
        position = position.open_volume if position is not None else 0
        best_bid, best_ask = vega.best_prices(market_id=market_id)
        return SimpleState(balance, position, best_bid, best_ask)

    def to_array(self) -> np.array:
        arr = np.nan_to_num(
            np.array([self.balance, self.position, self.best_bid, self.best_ask])
        )
        norm_factor = max(1e-12, np.linalg.norm(arr))
        arr = arr / norm_factor
        return arr


@dataclass(frozen=True)
class PriceStateWithFees(State):
    position: float
    full_balance: float
    market_in_auction: bool
    bid_prices: list[float]
    ask_prices: list[float]
    bid_volumes: list[float]
    ask_volumes: list[float]
    trading_fee: float

    def to_array(self) -> np.array:
        arr = np.nan_to_num(
            np.array(
                [
                    self.position,
                    self.full_balance,
                    int(self.market_in_auction),
                    self.trading_fee,
                ]
                + self.bid_prices
                + self.bid_volumes
                + self.ask_prices
                + self.ask_volumes
            )
        )
        norm_factor = max(1e-12, np.linalg.norm(arr))
        arr = arr / norm_factor
        return arr

    @classmethod
    def from_vega(
        cls,
        vega: VegaService,
        for_key_name: str,
        market_id: str,
        asset_id: str,
        for_wallet_name: Optional[str] = None,
    ) -> "PriceStateWithFees":
        position = vega.positions_by_market(
            for_key_name, market_id, wallet_name=for_wallet_name
        )

        position = position.open_volume if position else 0
        account = vega.party_account(
            key_name=for_key_name,
            asset_id=asset_id,
            market_id=market_id,
            wallet_name=for_wallet_name,
        )
        book_state = vega.market_depth(
            market_id, num_levels=5
        )  # make num_levels as a parameter?

        market_info = vega.market_info(market_id=market_id)
        fee = (
            float(market_info.fees.factors.liquidity_fee)
            + float(market_info.fees.factors.maker_fee)
            + float(market_info.fees.factors.infrastructure_fee)
        )

        bid_prices = [level.price for level in book_state.buys] + [0] * max(
            0, 5 - len(book_state.buys)
        )
        ask_prices = [level.price for level in book_state.sells] + [0] * max(
            0, 5 - len(book_state.sells)
        )

        return cls(
            position=position,
            full_balance=(account.margin + account.general),
            market_in_auction=(
                not market_info.trading_mode
                == markets_protos.Market.TradingMode.TRADING_MODE_CONTINUOUS
            ),
            bid_prices=bid_prices,
            ask_prices=ask_prices,
            bid_volumes=[level.volume for level in book_state.buys]
            + [0] * max(0, 5 - len(book_state.buys)),
            ask_volumes=[level.volume for level in book_state.sells]
            + [0] * max(0, 5 - len(book_state.sells)),
            trading_fee=fee,
        )


@dataclass(frozen=True)
class PositionOnly(State):
    position: float

    def to_array(self) -> np.array:
        return np.nan_to_num(
            np.array(
                [
                    self.position,
                ]
            )
        )

    @classmethod
    def from_vega(
        cls,
        vega: VegaService,
        for_key_name: str,
        market_id: str,
        asset_id: str,
        for_wallet_name: Optional[str] = None,
    ) -> "PositionOnly":
        position = vega.positions_by_market(
            for_key_name, market_id, wallet_name=for_wallet_name
        )

        return cls(
            position=position.open_volume if position else 0,
        )



from collections import namedtuple

PartyConfig = namedtuple("PartyConfig", ["wallet_name", "wallet_pass", "key_name"])

MARKET_CREATOR = PartyConfig("vega", "pass", "Market Creator")
MARKET_SETTLER = PartyConfig("vega", "pass", "Market Settler")

AUCTION_PASS_BID = PartyConfig("vega", "pass", "Auction Pass (Bid)")
AUCTION_PASS_ASK = PartyConfig("vega", "pass", "Auction Pass (Ask)")

INT_MARKET_MAKER_KEY_A = PartyConfig("vega", "pass", "Key A (Internal)")
INT_MARKET_MAKER_KEY_B = PartyConfig("vega", "pass", "Key B (External)")
INT_INFORMED_TRADER = PartyConfig("vega", "pass", "Informed Trader")
INT_RANDOM_TRADER = PartyConfig("vega", "pass", "Random Trader (Internal)")

EXT_MARKET_MAKER = PartyConfig("vega", "pass", "External Market Maker")
EXT_RANDOM_TRADER = PartyConfig("vega", "pass", "Random Trader (External)")
